# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
CPUID information

usage:
    >>> cpuid(0)
"""

from typing import Tuple
from chipsec.hal import hal_base
from chipsec.library.logger import logger


class CpuID(hal_base.HALBase):

    def __init__(self, cs):
        super(CpuID, self).__init__(cs)
        self.helper = cs.helper

    def cpuid(self, eax: int, ecx: int) -> Tuple[int, int, int, int]:
        logger().log_hal(f'[cpuid] in: EAX=0x{eax:08X}, ECX=0x{ecx:08X}')
        (eax, ebx, ecx, edx) = self.helper.cpuid(eax, ecx)
        logger().log_hal(f'[cpuid] out: EAX=0x{eax:08X}, EBX=0x{ebx:08X}, ECX=0x{ecx:08X}, EDX=0x{edx:08X}')
        return (eax, ebx, ecx, edx)

    def get_proc_info(self):
        (eax, _, _, _) = self.cpuid(0x01, 0x00)
        return eax



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

import struct
from typing import Dict


def DB(val: int) -> bytes:
    return struct.pack('<B', val)


def DW(val: int) -> bytes:
    return struct.pack('<H', val)


def DD(val: int) -> bytes:
    return struct.pack('<L', val)


def DQ(val: int) -> bytes:
    return struct.pack('<Q', val)


SIZE2FORMAT: Dict[int, str] = {
    1: 'B',
    2: 'H',
    4: 'I',
    8: 'Q'
}


def pack1(value: int, size: int) -> bytes:
    """Shortcut to pack a single value into a string based on its size."""
    return struct.pack(SIZE2FORMAT[size], value)


def unpack1(string: bytes, size: int) -> int:
    """Shortcut to unpack a single value from a string based on its size."""
    return struct.unpack(SIZE2FORMAT[size], string)[0]



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Common UEFI/EFI functionality including UEFI variables, Firmware Volumes, Secure Boot variables, S3 boot-script, UEFI tables, etc.
"""

import os
import struct
import codecs
from collections import namedtuple
from collections.abc import Callable
from typing import Dict, List, Tuple, Optional, Any
from uuid import UUID

from chipsec.library.file import read_file, write_file
from chipsec.library.logger import logger, dump_buffer, dump_buffer_bytes
from chipsec.library.defines import bytestostring

# from chipsec.helper.oshelper import helper


################################################################################################
#
# EFI Variable and Variable Store Defines
#
################################################################################################

# UDK2010.SR1\MdeModulePkg\Include\Guid\VariableFormat.h
#
# Variable data start flag.
#
VARIABLE_DATA = 0x55aa
VARIABLE_DATA_SIGNATURE = struct.pack('=H', VARIABLE_DATA)


#
# Variable Attributes
#
EFI_VARIABLE_NON_VOLATILE = 0x00000001  # Variable is non volatile
EFI_VARIABLE_BOOTSERVICE_ACCESS = 0x00000002  # Variable is boot time accessible
EFI_VARIABLE_RUNTIME_ACCESS = 0x00000004  # Variable is run-time accessible
EFI_VARIABLE_HARDWARE_ERROR_RECORD = 0x00000008
EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS = 0x00000010  # Variable is authenticated
EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS = 0x00000020  # Variable is time based authenticated
EFI_VARIABLE_APPEND_WRITE = 0x00000040  # Variable allows append
EFI_VARIABLE_ENHANCED_AUTHENTICATED_ACCESS = 0x00000080
UEFI23_1_AUTHENTICATED_VARIABLE_ATTRIBUTES = (EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS | EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS | EFI_VARIABLE_ENHANCED_AUTHENTICATED_ACCESS)


def IS_VARIABLE_ATTRIBUTE(_c: int, _Mask: int) -> bool:
    return ((_c & _Mask) != 0)


def IS_EFI_VARIABLE_AUTHENTICATED(attr: int) -> bool:
    return (IS_VARIABLE_ATTRIBUTE(attr, EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS) or
            IS_VARIABLE_ATTRIBUTE(attr, EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS) or
            IS_VARIABLE_ATTRIBUTE(attr, EFI_VARIABLE_ENHANCED_AUTHENTICATED_ACCESS))


MAX_VARIABLE_SIZE = 1024
MAX_NVRAM_SIZE = 1024 * 1024


def get_nvar_name(nvram: bytes, name_offset: int, isAscii: bool):
    if isAscii:
        nend = nvram.find(b'\x00', name_offset)
        name = nvram[name_offset:nend].decode('latin1')
        name_size = len(name) + 1
        return (name, name_size)
    else:
        nend = nvram.find(b'\x00\x00', name_offset)
        name = nvram[name_offset:nend].decode('utf-16le')
        name_size = len(name) + 2
        return (name, name_size)


VARIABLE_SIGNATURE_VSS = VARIABLE_DATA_SIGNATURE


VARIABLE_STORE_FV_GUID = UUID('FFF12B8D-7696-4C8B-A985-2747075B4F50')


################################################################################################
#
# Misc Defines
#
################################################################################################

#
# Status codes
# edk2: MdePkg/Include/Base.h
#

# @TODO
# define ENCODE_ERROR(StatusCode)     ((RETURN_STATUS)(MAX_BIT | (StatusCode)))
# define ENCODE_WARNING(a)            (a)

class StatusCode:
    EFI_SUCCESS = 0
    EFI_LOAD_ERROR = 1
    EFI_INVALID_PARAMETER = 2
    EFI_UNSUPPORTED = 3
    EFI_BAD_BUFFER_SIZE = 4
    EFI_BUFFER_TOO_SMALL = 5
    EFI_NOT_READY = 6
    EFI_DEVICE_ERROR = 7
    EFI_WRITE_PROTECTED = 8
    EFI_OUT_OF_RESOURCES = 9
    EFI_VOLUME_CORRUPTED = 10
    EFI_VOLUME_FULL = 11
    EFI_NO_MEDIA = 12
    EFI_MEDIA_CHANGED = 13
    EFI_NOT_FOUND = 14
    EFI_ACCESS_DENIED = 15
    EFI_NO_RESPONSE = 16
    EFI_NO_MAPPING = 17
    EFI_TIMEOUT = 18
    EFI_NOT_STARTED = 19
    EFI_ALREADY_STARTED = 20
    EFI_ABORTED = 21
    EFI_ICMP_ERROR = 22
    EFI_TFTP_ERROR = 23
    EFI_PROTOCOL_ERROR = 24
    EFI_INCOMPATIBLE_VERSION = 25
    EFI_SECURITY_VIOLATION = 26
    EFI_CRC_ERROR = 27
    EFI_END_OF_MEDIA = 28
    EFI_END_OF_FILE = 31
    EFI_INVALID_LANGUAGE = 32
    EFI_COMPROMISED_DATA = 33
    EFI_HTTP_ERROR = 35
    '''
  EFI_WARN_UNKNOWN_GLYPH    = 1
  EFI_WARN_DELETE_FAILURE   = 2
  EFI_WARN_WRITE_FAILURE    = 3
  EFI_WARN_BUFFER_TOO_SMALL = 4
  EFI_WARN_STALE_DATA       = 5
  EFI_WARN_FILE_SYSTEM      = 6
  '''


EFI_STATUS_DICT: Dict[int, str] = {
    StatusCode.EFI_SUCCESS: "EFI_SUCCESS",
    StatusCode.EFI_LOAD_ERROR: "EFI_LOAD_ERROR",
    StatusCode.EFI_INVALID_PARAMETER: "EFI_INVALID_PARAMETER",
    StatusCode.EFI_UNSUPPORTED: "EFI_UNSUPPORTED",
    StatusCode.EFI_BAD_BUFFER_SIZE: "EFI_BAD_BUFFER_SIZE",
    StatusCode.EFI_BUFFER_TOO_SMALL: "EFI_BUFFER_TOO_SMALL",
    StatusCode.EFI_NOT_READY: "EFI_NOT_READY",
    StatusCode.EFI_DEVICE_ERROR: "EFI_DEVICE_ERROR",
    StatusCode.EFI_WRITE_PROTECTED: "EFI_WRITE_PROTECTED",
    StatusCode.EFI_OUT_OF_RESOURCES: "EFI_OUT_OF_RESOURCES",
    StatusCode.EFI_VOLUME_CORRUPTED: "EFI_VOLUME_CORRUPTED",
    StatusCode.EFI_VOLUME_FULL: "EFI_VOLUME_FULL",
    StatusCode.EFI_NO_MEDIA: "EFI_NO_MEDIA",
    StatusCode.EFI_MEDIA_CHANGED: "EFI_MEDIA_CHANGED",
    StatusCode.EFI_NOT_FOUND: "EFI_NOT_FOUND",
    StatusCode.EFI_ACCESS_DENIED: "EFI_ACCESS_DENIED",
    StatusCode.EFI_NO_RESPONSE: "EFI_NO_RESPONSE",
    StatusCode.EFI_NO_MAPPING: "EFI_NO_MAPPING",
    StatusCode.EFI_TIMEOUT: "EFI_TIMEOUT",
    StatusCode.EFI_NOT_STARTED: "EFI_NOT_STARTED",
    StatusCode.EFI_ALREADY_STARTED: "EFI_ALREADY_STARTED",
    StatusCode.EFI_ABORTED: "EFI_ABORTED",
    StatusCode.EFI_ICMP_ERROR: "EFI_ICMP_ERROR",
    StatusCode.EFI_TFTP_ERROR: "EFI_TFTP_ERROR",
    StatusCode.EFI_PROTOCOL_ERROR: "EFI_PROTOCOL_ERROR",
    StatusCode.EFI_INCOMPATIBLE_VERSION: "EFI_INCOMPATIBLE_VERSION",
    StatusCode.EFI_SECURITY_VIOLATION: "EFI_SECURITY_VIOLATION",
    StatusCode.EFI_CRC_ERROR: "EFI_CRC_ERROR",
    StatusCode.EFI_END_OF_MEDIA: "EFI_END_OF_MEDIA",
    StatusCode.EFI_END_OF_FILE: "EFI_END_OF_FILE",
    StatusCode.EFI_INVALID_LANGUAGE: "EFI_INVALID_LANGUAGE",
    StatusCode.EFI_COMPROMISED_DATA: "EFI_COMPROMISED_DATA",
    StatusCode.EFI_HTTP_ERROR: "EFI_HTTP_ERROR"
}

EFI_MAX_BIT = 0x8000000000000000


def EFI_ERROR_STR(error: int) -> str:
    """
    Translates an EFI_STATUS value into its corresponding textual representation.
    """
    error &= ~EFI_MAX_BIT
    try:
        return EFI_STATUS_DICT[error]
    except KeyError:
        return "UNKNOWN"


EFI_GUID_FMT = "16s"
EFI_GUID_SIZE = struct.calcsize(EFI_GUID_FMT)


def EFI_GUID_STR(guid: bytes) -> str:
    guid_str = UUID(bytes_le=guid)
    return str(guid_str).upper()


def align(of: int, size: int) -> int:
    of = (((of + size - 1) // size) * size)
    return of


def bit_set(value: int, mask: int, polarity: bool = False) -> bool:
    if polarity:
        value = ~value
    return ((value & mask) == mask)


def get_3b_size(s_data: bytes) -> int:
    s_str = bytestostring(s_data)
    return (ord(s_str[0]) + (ord(s_str[1]) << 8) + (ord(s_str[2]) << 16))


# #################################################################################################
#
# UEFI Variable (NVRAM) Parsing Functionality
#
# #################################################################################################

SIGNATURE_LIST = "<16sIII"
SIGNATURE_LIST_size = struct.calcsize(SIGNATURE_LIST)


def parse_sha256(data):
    return


def parse_rsa2048(data):
    return


def parse_rsa2048_sha256(data):
    return


def parse_sha1(data):
    return


def parse_rsa2048_sha1(data):
    return


def parse_x509(data):
    return


def parse_sha224(data):
    return


def parse_sha384(data):
    return


def parse_sha512(data):
    return


def parse_x509_sha256(data):
    return


def parse_x509_sha384(data):
    return


def parse_x509_sha512(data):
    return


def parse_external(data):
    return


def parse_pkcs7(data):
    return


sig_types: Dict[str, Tuple[str, Callable, int, str]] = {
    "C1C41626-504C-4092-ACA9-41F936934328": ("EFI_CERT_SHA256_GUID", parse_sha256, 0x30, "SHA256"),
    "3C5766E8-269C-4E34-AA14-ED776E85B3B6": ("EFI_CERT_RSA2048_GUID", parse_rsa2048, 0x110, "RSA2048"),
    "E2B36190-879B-4A3D-AD8D-F2E7BBA32784": ("EFI_CERT_RSA2048_SHA256_GUID", parse_rsa2048_sha256, 0x110, "RSA2048_SHA256"),
    "826CA512-CF10-4AC9-B187-BE01496631BD": ("EFI_CERT_SHA1_GUID", parse_sha1, 0x24, "SHA1"),
    "67F8444F-8743-48F1-A328-1EAAB8736080": ("EFI_CERT_RSA2048_SHA1_GUID", parse_rsa2048_sha1, 0x110, "RSA2048_SHA1"),
    "A5C059A1-94E4-4AA7-87B5-AB155C2BF072": ("EFI_CERT_X509_GUID", parse_x509, 0, "X509"),
    "0B6E5233-A65C-44C9-9407-D9AB83BFC8BD": ("EFI_CERT_SHA224_GUID", parse_sha224, 0x2c, "SHA224"),
    "FF3E5307-9FD0-48C9-85F1-8AD56C701E01": ("EFI_CERT_SHA384_GUID", parse_sha384, 0x40, "SHA384"),
    "093E0FAE-A6C4-4F50-9F1B-D41E2B89C19A": ("EFI_CERT_SHA512_GUID", parse_sha512, 0x50, "SHA512"),
    "3bd2a492-96c0-4079-b420-fcf98ef103ed": ("EFI_CERT_X509_SHA256_GUID", parse_x509_sha256, 0x40, "X509_SHA256"),
    "7076876e-80c2-4ee6-aad2-28b349a6865b": ("EFI_CERT_X509_SHA384_GUID", parse_x509_sha384, 0x50, "X509_SHA384"),
    "446dbf63-2502-4cda-bcfa-2465d2b0fe9d": ("EFI_CERT_X509_SHA512_GUID", parse_x509_sha512, 0x60, "X509_SHA512"),
    "452e8ced-dfff-4b8c-ae01-5118862e682c": ("EFI_CERT_EXTERNAL_MANAGEMENT_GUID", parse_external, 0x11, "EXTERNAL_MANAGEMENT"),
    "4AAFD29D-68DF-49EE-8AA9-347D375665A7": ("EFI_CERT_TYPE_PKCS7_GUID", parse_pkcs7, 0, "PKCS7"),
}


def parse_sb_db(db: bytes, decode_dir: str) -> List[bytes]:
    entries = []
    dof = 0
    nsig = 0
    db_size = len(db)
    if 0 == db_size:
        return entries

    # some platforms have 0's in the beginnig, skip all 0 (no known SignatureType starts with 0x00):
    while (dof + SIGNATURE_LIST_size) < db_size:
        SignatureType0, SignatureListSize, SignatureHeaderSize, SignatureSize \
            = struct.unpack(SIGNATURE_LIST, db[dof:dof + SIGNATURE_LIST_size])

        # prevent infinite loop when parsing malformed var
        if SignatureListSize == 0:
            logger().log_bad("db parsing failed!")
            return entries

        # Determine the signature type
        SignatureType = EFI_GUID_STR(SignatureType0)
        sig_parse_f = None
        sig_size = 0
        if (SignatureType in sig_types.keys()):
            sig_name, sig_parse_f, sig_size, short_name = sig_types[SignatureType]
        else:
            logger().log_bad(f'Unknown signature type {SignatureType}, skipping signature decode.')
            dof += SignatureListSize
            continue

        # Extract signature data blobs
        if (((sig_size > 0) and (sig_size == SignatureSize)) or ((sig_size == 0) and (SignatureSize >= 0x10))):
            sof = 0
            sig_list = db[dof + SIGNATURE_LIST_size + SignatureHeaderSize:dof + SignatureListSize]
            sig_list_size = len(sig_list)
            while ((sof + EFI_GUID_SIZE) < sig_list_size):
                sig_data = sig_list[sof:sof + SignatureSize]
                owner0 = struct.unpack(EFI_GUID_FMT, sig_data[:EFI_GUID_SIZE])[0]
                owner = EFI_GUID_STR(owner0)
                data = sig_data[EFI_GUID_SIZE:]
                entries.append(data)
                sig_file_name = f'{short_name}-{owner}-{nsig:02d}.bin'
                sig_file_name = os.path.join(decode_dir, sig_file_name)
                write_file(sig_file_name, data)
                if (sig_parse_f is not None):
                    sig_parse_f(data)
                sof = sof + SignatureSize
                nsig = nsig + 1
        else:
            err_str = f'Wrong SignatureSize for {SignatureType} type: 0x{SignatureSize:X}.'
            if (sig_size > 0):
                err_str = err_str + f' Must be 0x{sig_size:X}.'
            else:
                err_str = err_str + " Must be >= 0x10."
            logger().log_error(err_str)
            logger().log_error('Skipping signature decode for this list.')
        dof = dof + SignatureListSize

    return entries


#
#  "certdb" variable stores the signer's certificates for non PK/KEK/DB/DBX
# variables with EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS|EFI_VARIABLE_NON_VOLATILE set.
#  "certdbv" variable stores the signer's certificates for non PK/KEK/DB/DBX
# variables with EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS set
#
# GUID: gEfiCertDbGuid
#
# We need maintain atomicity.
#
# Format:
# +----------------------------+
# | UINT32                     | <-- CertDbListSize, including this UINT32
# +----------------------------+
# | AUTH_CERT_DB_DATA          | <-- First CERT
# +----------------------------+
# | ........                   |
# +----------------------------+
# | AUTH_CERT_DB_DATA          | <-- Last CERT
# +----------------------------+
#
# typedef struct {
#   EFI_GUID    VendorGuid;
#   UINT32      CertNodeSize;
#   UINT32      NameSize;
#   UINT32      CertDataSize;
#   /// CHAR16  VariableName[NameSize];
#   /// UINT8   CertData[CertDataSize];
# } AUTH_CERT_DB_DATA;
#
AUTH_CERT_DB_LIST_HEAD = "<I"
AUTH_CERT_DB_LIST_HEAD_size = struct.calcsize(AUTH_CERT_DB_LIST_HEAD)
AUTH_CERT_DB_DATA = "<16sIII"
AUTH_CERT_DB_DATA_size = struct.calcsize(AUTH_CERT_DB_DATA)


def parse_auth_var(db: bytes, decode_dir: str) -> List[bytes]:
    entries = []
    dof = 0
    nsig = 0
    db_size = len(db)

    # Verify that list makes sense
    if db_size < AUTH_CERT_DB_LIST_HEAD_size:
        logger().log_warning("Cert list empty.")
        return entries
    expected_size = struct.unpack(AUTH_CERT_DB_LIST_HEAD, db[dof:dof + AUTH_CERT_DB_LIST_HEAD_size])[0]
    if db_size != expected_size:
        logger().log_error("Expected size of cert list did not match actual size.")
        return entries
    dof += AUTH_CERT_DB_LIST_HEAD_size

    # Loop through all the certs in the list.
    while dof + AUTH_CERT_DB_DATA_size < db_size:
        ven_guid0, cert_node_size, name_size, cert_data_size = struct.unpack(AUTH_CERT_DB_DATA, db[dof:dof + AUTH_CERT_DB_DATA_size])
        vendor_guid = EFI_GUID_STR(ven_guid0)
        name_size *= 2  # Name size is actually the number of CHAR16 in the name array
        tof = dof + AUTH_CERT_DB_DATA_size
        try:
            var_name = codecs.decode(db[tof:tof + name_size], 'utf-16')
        except UnicodeDecodeError:
            logger().log_warning(f'Unable to decode {db[tof:tof + name_size]}')
            var_name = "chipsec.library.exceptions!"
        tof += name_size
        sig_data = db[tof:tof + cert_data_size]
        entries.append(sig_data)
        sig_file_name = f'{vendor_guid}-{codecs.encode(var_name)}-{nsig:02X}.bin'
        sig_file_name = os.path.join(decode_dir, sig_file_name)
        write_file(sig_file_name, sig_data)
        dof += cert_node_size
        nsig += 1

    return entries


ESAL_SIG_SIZE = 256


def parse_esal_var(db: bytes, decode_dir: str) -> List[bytes]:
    entries = []
    dof = 0
    nsig = 0
    db_size = len(db)

    # Check to see how many signatures exist
    if db_size < ESAL_SIG_SIZE:
        logger().log('No signatures present.')
        return entries

    # Extract signatures
    while dof + ESAL_SIG_SIZE <= db_size:
        key_data = db[dof:dof + ESAL_SIG_SIZE]
        entries.append(key_data)
        key_file_name = os.path.join(decode_dir, f'AuthVarKeyDatabase-cert-{nsig:02X}.bin')
        write_file(key_file_name, key_data)
        dof += ESAL_SIG_SIZE
        nsig += 1

    return entries


SECURE_BOOT_SIG_VAR = 1
AUTH_SIG_VAR = 2
ESAL_SIG_VAR = 3


def parse_efivar_file(fname: str, var: Optional[bytes] = None, var_type: int = SECURE_BOOT_SIG_VAR) -> None:
    logger().log(f'Processing certs in file: {fname}')
    if not var:
        var = read_file(fname)
    var_path = fname + '.dir'
    if not os.path.exists(var_path):
        os.makedirs(var_path)
    if var_type == SECURE_BOOT_SIG_VAR:
        parse_sb_db(var, var_path)
    elif var_type == AUTH_SIG_VAR:
        parse_auth_var(var, var_path)
    elif var_type == ESAL_SIG_VAR:
        parse_esal_var(var, var_path)
    else:
        logger().log_warning(f'Unsupported variable type requested: {var_type}')


########################################################################################################
#
# S3 Resume Boot-Script Parsing Functionality
#
########################################################################################################

BOOTSCRIPT_TABLE_OFFSET = 24
RUNTIME_SCRIPT_TABLE_BASE_OFFSET = 32
ACPI_VARIABLE_SET_STRUCT_SIZE = 0x48
S3_BOOTSCRIPT_VARIABLES = ['AcpiGlobalVariable']

MAX_S3_BOOTSCRIPT_ENTRY_LENGTH = 0x200


#
# MdePkg\Include\Pi\PiS3BootScript.h
#
# //*******************************************
# // EFI Boot Script Opcode definitions
# //*******************************************

class S3BootScriptOpcode:
    EFI_BOOT_SCRIPT_IO_WRITE_OPCODE = 0x00
    EFI_BOOT_SCRIPT_IO_READ_WRITE_OPCODE = 0x01
    EFI_BOOT_SCRIPT_MEM_WRITE_OPCODE = 0x02
    EFI_BOOT_SCRIPT_MEM_READ_WRITE_OPCODE = 0x03
    EFI_BOOT_SCRIPT_PCI_CONFIG_WRITE_OPCODE = 0x04
    EFI_BOOT_SCRIPT_PCI_CONFIG_READ_WRITE_OPCODE = 0x05
    EFI_BOOT_SCRIPT_SMBUS_EXECUTE_OPCODE = 0x06
    EFI_BOOT_SCRIPT_STALL_OPCODE = 0x07
    EFI_BOOT_SCRIPT_DISPATCH_OPCODE = 0x08
    EFI_BOOT_SCRIPT_TERMINATE_OPCODE = 0xFF


class S3BootScriptOpcode_MDE (S3BootScriptOpcode):
    EFI_BOOT_SCRIPT_DISPATCH_2_OPCODE = 0x09
    EFI_BOOT_SCRIPT_INFORMATION_OPCODE = 0x0A
    EFI_BOOT_SCRIPT_PCI_CONFIG2_WRITE_OPCODE = 0x0B
    EFI_BOOT_SCRIPT_PCI_CONFIG2_READ_WRITE_OPCODE = 0x0C
    EFI_BOOT_SCRIPT_IO_POLL_OPCODE = 0x0D
    EFI_BOOT_SCRIPT_MEM_POLL_OPCODE = 0x0E
    EFI_BOOT_SCRIPT_PCI_CONFIG_POLL_OPCODE = 0x0F
    EFI_BOOT_SCRIPT_PCI_CONFIG2_POLL_OPCODE = 0x10

#
# EdkCompatibilityPkg\Foundation\Framework\Include\EfiBootScript.h
#


class S3BootScriptOpcode_EdkCompat (S3BootScriptOpcode):
    EFI_BOOT_SCRIPT_MEM_POLL_OPCODE = 0x09
    EFI_BOOT_SCRIPT_INFORMATION_OPCODE = 0x0A
    EFI_BOOT_SCRIPT_PCI_CONFIG2_WRITE_OPCODE = 0x0B
    EFI_BOOT_SCRIPT_PCI_CONFIG2_READ_WRITE_OPCODE = 0x0C
    EFI_BOOT_SCRIPT_TABLE_OPCODE = 0xAA


#
# Names of S3 Boot Script Opcodes
#
script_opcodes: Dict[int, str] = {
    S3BootScriptOpcode.EFI_BOOT_SCRIPT_IO_WRITE_OPCODE: "S3_BOOTSCRIPT_IO_WRITE",
    S3BootScriptOpcode.EFI_BOOT_SCRIPT_IO_READ_WRITE_OPCODE: "S3_BOOTSCRIPT_IO_READ_WRITE",
    S3BootScriptOpcode.EFI_BOOT_SCRIPT_MEM_WRITE_OPCODE: "S3_BOOTSCRIPT_MEM_WRITE",
    S3BootScriptOpcode.EFI_BOOT_SCRIPT_MEM_READ_WRITE_OPCODE: "S3_BOOTSCRIPT_MEM_READ_WRITE",
    S3BootScriptOpcode.EFI_BOOT_SCRIPT_PCI_CONFIG_WRITE_OPCODE: "S3_BOOTSCRIPT_PCI_CONFIG_WRITE",
    S3BootScriptOpcode.EFI_BOOT_SCRIPT_PCI_CONFIG_READ_WRITE_OPCODE: "S3_BOOTSCRIPT_PCI_CONFIG_READ_WRITE",
    S3BootScriptOpcode.EFI_BOOT_SCRIPT_SMBUS_EXECUTE_OPCODE: "S3_BOOTSCRIPT_SMBUS_EXECUTE",
    S3BootScriptOpcode.EFI_BOOT_SCRIPT_STALL_OPCODE: "S3_BOOTSCRIPT_STALL",
    S3BootScriptOpcode.EFI_BOOT_SCRIPT_DISPATCH_OPCODE: "S3_BOOTSCRIPT_DISPATCH",
    # S3BootScriptOpcode.EFI_BOOT_SCRIPT_DISPATCH_2_OPCODE:             "S3_BOOTSCRIPT_DISPATCH_2",
    # S3BootScriptOpcode.EFI_BOOT_SCRIPT_INFORMATION_OPCODE:            "S3_BOOTSCRIPT_INFORMATION",
    # S3BootScriptOpcode.EFI_BOOT_SCRIPT_PCI_CONFIG2_WRITE_OPCODE:      "S3_BOOTSCRIPT_PCI_CONFIG2_WRITE",
    # S3BootScriptOpcode.EFI_BOOT_SCRIPT_PCI_CONFIG2_READ_WRITE_OPCODE: "S3_BOOTSCRIPT_PCI_CONFIG2_READ_WRITE",
    # S3BootScriptOpcode.EFI_BOOT_SCRIPT_IO_POLL_OPCODE:                "S3_BOOTSCRIPT_IO_POLL",
    # S3BootScriptOpcode.EFI_BOOT_SCRIPT_MEM_POLL_OPCODE:               "S3_BOOTSCRIPT_MEM_POLL",
    # S3BootScriptOpcode.EFI_BOOT_SCRIPT_PCI_CONFIG_POLL_OPCODE:        "S3_BOOTSCRIPT_PCI_CONFIG_POLL",
    # S3BootScriptOpcode.EFI_BOOT_SCRIPT_PCI_CONFIG2_POLL_OPCODE:       "S3_BOOTSCRIPT_PCI_CONFIG2_POLL",
    # S3BootScriptOpcode.EFI_BOOT_SCRIPT_TABLE_OPCODE:                  "S3_BOOTSCRIPT_TABLE",
    S3BootScriptOpcode.EFI_BOOT_SCRIPT_TERMINATE_OPCODE: "S3_BOOTSCRIPT_TERMINATE"
}


# //*******************************************
# // EFI_BOOT_SCRIPT_WIDTH
# //*******************************************
# typedef enum {
# EfiBootScriptWidthUint8,
# EfiBootScriptWidthUint16,
# EfiBootScriptWidthUint32,
# EfiBootScriptWidthUint64,
# EfiBootScriptWidthFifoUint8,
# EfiBootScriptWidthFifoUint16,
# EfiBootScriptWidthFifoUint32,
# EfiBootScriptWidthFifoUint64,
# EfiBootScriptWidthFillUint8,
# EfiBootScriptWidthFillUint16,
# EfiBootScriptWidthFillUint32,
# EfiBootScriptWidthFillUint64,
# EfiBootScriptWidthMaximum
# } EFI_BOOT_SCRIPT_WIDTH;

class S3BootScriptWidth:
    EFI_BOOT_SCRIPT_WIDTH_UINT8 = 0x00
    EFI_BOOT_SCRIPT_WIDTH_UINT16 = 0x01
    EFI_BOOT_SCRIPT_WIDTH_UINT32 = 0x02
    EFI_BOOT_SCRIPT_WIDTH_UINT64 = 0x03


script_width_sizes: Dict[int, int] = {
    S3BootScriptWidth.EFI_BOOT_SCRIPT_WIDTH_UINT8: 1,
    S3BootScriptWidth.EFI_BOOT_SCRIPT_WIDTH_UINT16: 2,
    S3BootScriptWidth.EFI_BOOT_SCRIPT_WIDTH_UINT32: 4,
    S3BootScriptWidth.EFI_BOOT_SCRIPT_WIDTH_UINT64: 8
}

script_width_values: Dict[int, int] = {
    1: S3BootScriptWidth.EFI_BOOT_SCRIPT_WIDTH_UINT8,
    2: S3BootScriptWidth.EFI_BOOT_SCRIPT_WIDTH_UINT16,
    4: S3BootScriptWidth.EFI_BOOT_SCRIPT_WIDTH_UINT32,
    8: S3BootScriptWidth.EFI_BOOT_SCRIPT_WIDTH_UINT64
}

script_width_formats: Dict[int, str] = {
    S3BootScriptWidth.EFI_BOOT_SCRIPT_WIDTH_UINT8: 'B',
    S3BootScriptWidth.EFI_BOOT_SCRIPT_WIDTH_UINT16: 'H',
    S3BootScriptWidth.EFI_BOOT_SCRIPT_WIDTH_UINT32: 'I',
    S3BootScriptWidth.EFI_BOOT_SCRIPT_WIDTH_UINT64: 'Q'
}

# //************************************************
# // EFI_SMBUS_DEVICE_ADDRESS
# //************************************************
# typedef struct _EFI_SMBUS_DEVICE_ADDRESS {
# UINTN SmbusDeviceAddress:7;
# } EFI_SMBUS_DEVICE_ADDRESS;
# //************************************************
# // EFI_SMBUS_DEVICE_COMMAND
# //************************************************
# typedef UINTN EFI_SMBUS_DEVICE_COMMAND;
#
# //************************************************
# // EFI_SMBUS_OPERATION
# //************************************************
# typedef enum _EFI_SMBUS_OPERATION {
# EfiSmbusQuickRead,
# EfiSmbusQuickWrite,
# EfiSmbusReceiveByte,
# EfiSmbusSendByte,
# EfiSmbusReadByte,
# EfiSmbusWriteByte,
# EfiSmbusReadWord,
# EfiSmbusWriteWord,
# EfiSmbusReadBlock,
# EfiSmbusWriteBlock,
# EfiSmbusProcessCall,
# EfiSmbusBWBRProcessCall
# } EFI_SMBUS_OPERATION;


class S3BootScriptSmbusOperation:
    QUICK_READ = 0x00
    QUICK_WRITE = 0x01
    RECEIVE_BYTE = 0x02
    SEND_BYTE = 0x03
    READ_BYTE = 0x04
    WRITE_BYTE = 0x05
    READ_WORD = 0x06
    WRITE_WORD = 0x07
    READ_BLOCK = 0x08
    WRITE_BLOCK = 0x09
    PROCESS_CALL = 0x0A
    BWBR_PROCESS_CALL = 0x0B


class op_io_pci_mem:
    def __init__(self, opcode: int, size: int, width: int, address: int, unknown: Optional[int], count: Optional[int],
                 buffer: Optional[bytes], value: Optional[int] = None, mask: Optional[int] = None):
        self.opcode = opcode
        self.size = size
        self.width = width
        self.address = address
        self.unknown = unknown
        self.count = count
        self.value = value
        self.mask = mask
        self.name = script_opcodes[opcode]
        self.buffer = buffer  # data[ self.size : ]
        self.values = None
        if self.count is not None and self.count > 0 and self.buffer is not None:
            sz = self.count * script_width_sizes[self.width]
            if len(self.buffer) != sz:
                logger().log(f'[?] buffer size (0x{len(self.buffer):X}) != Width x Count (0x{sz:X})')
            else:
                self.values = list(struct.unpack((f'<{self.count:d}{script_width_formats[self.width]:1}'), self.buffer))

    def __str__(self) -> str:
        str_r = f'  Opcode : {self.name} (0x{self.opcode:04X})\n'
        str_r += f'  Width  : 0x{self.width:02X} ({script_width_sizes[self.width]:X} bytes)\n'
        str_r += f'  Address: 0x{self.address:08X}\n'
        if self.value is not None:
            str_r += f'  Value  : 0x{self.value:08X}\n'
        if self.mask is not None:
            str_r += f'  Mask   : 0x{self.mask:08X}\n'
        if self.unknown is not None:
            str_r += f'  Unknown: 0x{self.unknown:04X}\n'
        if self.count is not None:
            str_r += f'  Count  : 0x{self.count:X}\n'
        if self.values is not None:
            fmt = f'0x{{:0{script_width_sizes[self.width] * 2:d}X}}'
            values_str = '  '.join([fmt.format(v) for v in self.values])
            str_r += f'  Values : {values_str}\n'
        elif self.buffer is not None:
            str_r += f'  Buffer (size = 0x{len(self.buffer):X}):\n{dump_buffer(self.buffer, 16)}'
        return str_r


class op_smbus_execute:
    def __init__(self, opcode: int, size: int, address: int, command: int, operation: int, peccheck: int):
        self.opcode = opcode
        self.size = size
        self.address = address
        self.command = command
        self.operation = operation
        self.peccheck = peccheck
        self.name = script_opcodes[opcode]

    def __str__(self) -> str:
        str_r = f'  Opcode           : {self.name} (0x{self.opcode:04X})\n'
        str_r += f'  Secondary Address: 0x{self.address:02X}\n'
        str_r += f'  Command          : 0x{self.command:08X}\n'
        str_r += f'  Operation        : 0x{self.operation:02X}\n'
        str_r += f'  PEC Check        : {self.peccheck:d}\n'
        return str_r

# typedef struct {
#  UINT16  OpCode;
#  UINT8   Length;
#  UINT64  Duration;
# } EFI_BOOT_SCRIPT_STALL;


class op_stall:
    def __init__(self, opcode: int, size: int, duration: int):
        self.opcode = opcode
        self.size = size
        self.duration = duration
        self.name = script_opcodes[self.opcode]

    def __str__(self) -> str:
        str_r = f'  Opcode  : {self.name} (0x{self.opcode:04X})\n'
        str_r += f'  Duration: 0x{self.duration:08X} (us)\n'
        return str_r

# typedef struct {
#  UINT16                OpCode;
#  UINT8                 Length;
#  EFI_PHYSICAL_ADDRESS  EntryPoint;
# } EFI_BOOT_SCRIPT_DISPATCH;


class op_dispatch:
    def __init__(self, opcode: int, size: int, entrypoint: int, context: Optional[int] = None):
        self.opcode = opcode
        self.size = size
        self.entrypoint = entrypoint
        self.context = context
        self.name = script_opcodes[self.opcode]

    def __str__(self) -> str:
        str_r = f'  Opcode     : {self.name} (0x{self.opcode:04X})\n'
        str_r += f'  Entry Point: 0x{self.entrypoint:016X}\n'
        if self.context is not None:
            str_r += f'  Context    : 0x{self.context:016X}\n'
        return str_r

# typedef struct {
#  UINT16  OpCode;
#  UINT8   Length;
#  UINT32  Width;
#  UINT64  Address;
#  UINT64  Duration;
#  UINT64  LoopTimes;
# } EFI_BOOT_SCRIPT_MEM_POLL;


class op_mem_poll:
    def __init__(self, opcode: int, size: int, width: int, address: int, duration: int, looptimes: int):
        self.opcode = opcode
        self.size = size
        self.width = width
        self.address = address
        self.duration = duration
        self.looptimes = looptimes
        self.name = 'S3_BOOTSCRIPT_MEM_POLL'

    def __str__(self) -> str:
        str_r = f'  Opcode    : {self.name} (0x{self.opcode:04X})\n'
        str_r += f'  Width     : 0x{self.width:02X} ({script_width_sizes[self.width]:X} bytes)\n'
        str_r += f'  Address   : 0x{self.address:016X}\n'
        str_r += f'  Duration? : 0x{self.duration:016X}\n'
        str_r += f'  LoopTimes?: 0x{self.looptimes:016X}\n'
        return str_r


class op_terminate:
    def __init__(self, opcode: int, size: int):
        self.opcode = opcode
        self.size = size
        self.name = script_opcodes[self.opcode]

    def __str__(self) -> str:
        return f'  Opcode     : {self.name} (0x{self.opcode:02X})\n'


class op_unknown:
    def __init__(self, opcode: int, size: int):
        self.opcode = opcode
        self.size = size

    def __str__(self) -> str:
        return f'  Opcode     : unknown (0x{self.opcode:02X})\n'


class S3BOOTSCRIPT_ENTRY:
    def __init__(self, script_type: int, index: Optional[int], offset_in_script: int, length: int, data: Optional[bytes] = None):
        self.script_type = script_type
        self.index = index
        self.offset_in_script = offset_in_script
        self.length = length
        self.data = data
        self.decoded_opcode = None
        self.header_length = 0

    def __str__(self) -> str:
        entry_str = '' if self.index is None else (f'[{self.index:03d}] ')
        entry_str += f'Entry at offset 0x{self.offset_in_script:04X} (len = 0x{self.length:X}, header len = 0x{self.header_length:X}):'
        if self.data:
            entry_str = entry_str + f'\nData:\n{dump_buffer_bytes(self.data, 16)}'
        if self.decoded_opcode:
            entry_str = entry_str + f'Decoded:\n{str(self.decoded_opcode)}'
        return entry_str


# #################################################################################################
#
# UEFI Table Parsing Functionality
#
# #################################################################################################
MAX_EFI_TABLE_SIZE = 0x1000

# typedef struct {
#   UINT64  Signature;
#   UINT32  Revision;
#   UINT32  HeaderSize;
#   UINT32  CRC32;
#   UINT32  Reserved;
# } EFI_TABLE_HEADER;

EFI_TABLE_HEADER_FMT = '=8sIIII'
EFI_TABLE_HEADER_SIZE = 0x18


class EFI_TABLE_HEADER(namedtuple('EFI_TABLE_HEADER', 'Signature Revision HeaderSize CRC32 Reserved')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""Header:
  Signature     : {bytestostring(self.Signature)}
  Revision      : {EFI_SYSTEM_TABLE_REVISION(self.Revision)}
  HeaderSize    : 0x{self.HeaderSize:08X}
  CRC32         : 0x{self.CRC32:08X}
  Reserved      : 0x{self.Reserved:08X}"""


# #################################################################################################
# EFI System Table
# #################################################################################################
#
# \MdePkg\Include\Uefi\UefiSpec.h

EFI_SYSTEM_TABLE_SIGNATURE = 'IBI SYST'

EFI_2_80_SYSTEM_TABLE_REVISION = ((2 << 16) | (80))
EFI_2_70_SYSTEM_TABLE_REVISION = ((2 << 16) | (70))
EFI_2_60_SYSTEM_TABLE_REVISION = ((2 << 16) | (60))
EFI_2_50_SYSTEM_TABLE_REVISION = ((2 << 16) | (50))
EFI_2_40_SYSTEM_TABLE_REVISION = ((2 << 16) | (40))
EFI_2_31_SYSTEM_TABLE_REVISION = ((2 << 16) | (31))
EFI_2_30_SYSTEM_TABLE_REVISION = ((2 << 16) | (30))
EFI_2_20_SYSTEM_TABLE_REVISION = ((2 << 16) | (20))
EFI_2_10_SYSTEM_TABLE_REVISION = ((2 << 16) | (10))
EFI_2_00_SYSTEM_TABLE_REVISION = ((2 << 16) | (00))
EFI_1_10_SYSTEM_TABLE_REVISION = ((1 << 16) | (10))
EFI_1_02_SYSTEM_TABLE_REVISION = ((1 << 16) | (0o2))
EFI_REVISIONS: List[int] = [
    EFI_2_80_SYSTEM_TABLE_REVISION,
    EFI_2_70_SYSTEM_TABLE_REVISION,
    EFI_2_60_SYSTEM_TABLE_REVISION,
    EFI_2_50_SYSTEM_TABLE_REVISION,
    EFI_2_40_SYSTEM_TABLE_REVISION,
    EFI_2_31_SYSTEM_TABLE_REVISION,
    EFI_2_30_SYSTEM_TABLE_REVISION,
    EFI_2_20_SYSTEM_TABLE_REVISION,
    EFI_2_10_SYSTEM_TABLE_REVISION,
    EFI_2_00_SYSTEM_TABLE_REVISION,
    EFI_1_10_SYSTEM_TABLE_REVISION,
    EFI_1_02_SYSTEM_TABLE_REVISION
]


def EFI_SYSTEM_TABLE_REVISION(revision: int) -> str:
    return f'{revision >> 16:d}.{revision & 0xFFFF:d}'


EFI_SYSTEM_TABLE_FMT = '=12Q'


class EFI_SYSTEM_TABLE(namedtuple('EFI_SYSTEM_TABLE', 'FirmwareVendor FirmwareRevision ConsoleInHandle ConIn ConsoleOutHandle ConOut StandardErrorHandle StdErr RuntimeServices BootServices NumberOfTableEntries ConfigurationTable')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""EFI System Table:
  FirmwareVendor      : 0x{self.FirmwareVendor:016X}
  FirmwareRevision    : 0x{self.FirmwareRevision:016X}
  ConsoleInHandle     : 0x{self.ConsoleInHandle:016X}
  ConIn               : 0x{self.ConIn:016X}
  ConsoleOutHandle    : 0x{self.ConsoleOutHandle:016X}
  ConOut              : 0x{self.ConOut:016X}
  StandardErrorHandle : 0x{self.StandardErrorHandle:016X}
  StdErr              : 0x{self.StdErr:016X}
  RuntimeServices     : 0x{self.RuntimeServices:016X}
  BootServices        : 0x{self.BootServices:016X}
  NumberOfTableEntries: 0x{self.NumberOfTableEntries:016X}
  ConfigurationTable  : 0x{self.ConfigurationTable:016X}
"""


# #################################################################################################
# EFI Runtime Services Table
# #################################################################################################
#
# \MdePkg\Include\Uefi\UefiSpec.h

EFI_RUNTIME_SERVICES_SIGNATURE = 'RUNTSERV'
EFI_RUNTIME_SERVICES_REVISION = EFI_2_31_SYSTEM_TABLE_REVISION

EFI_RUNTIME_SERVICES_TABLE_FMT = '=14Q'


class EFI_RUNTIME_SERVICES_TABLE(namedtuple('EFI_RUNTIME_SERVICES_TABLE', 'GetTime SetTime GetWakeupTime SetWakeupTime SetVirtualAddressMap ConvertPointer GetVariable GetNextVariableName SetVariable GetNextHighMonotonicCount ResetSystem UpdateCapsule QueryCapsuleCapabilities QueryVariableInfo')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""Runtime Services:
  GetTime                  : 0x{self.GetTime:016X}
  SetTime                  : 0x{self.SetTime:016X}
  GetWakeupTime            : 0x{self.GetWakeupTime:016X}
  SetWakeupTime            : 0x{self.SetWakeupTime:016X}
  SetVirtualAddressMap     : 0x{self.SetVirtualAddressMap:016X}
  ConvertPointer           : 0x{self.ConvertPointer:016X}
  GetVariable              : 0x{self.GetVariable:016X}
  GetNextVariableName      : 0x{self.GetNextVariableName:016X}
  SetVariable              : 0x{self.SetVariable:016X}
  GetNextHighMonotonicCount: 0x{self.GetNextHighMonotonicCount:016X}
  ResetSystem              : 0x{self.ResetSystem:016X}
  UpdateCapsule            : 0x{self.UpdateCapsule:016X}
  QueryCapsuleCapabilities : 0x{self.QueryCapsuleCapabilities:016X}
  QueryVariableInfo        : 0x{self.QueryVariableInfo:016X}
"""


# #################################################################################################
# EFI Boot Services Table
# #################################################################################################
#
# \MdePkg\Include\Uefi\UefiSpec.h

EFI_BOOT_SERVICES_SIGNATURE = 'BOOTSERV'
EFI_BOOT_SERVICES_REVISION = EFI_2_31_SYSTEM_TABLE_REVISION

EFI_BOOT_SERVICES_TABLE_FMT = '=44Q'


class EFI_BOOT_SERVICES_TABLE(namedtuple('EFI_BOOT_SERVICES_TABLE', 'RaiseTPL RestoreTPL AllocatePages FreePages GetMemoryMap AllocatePool FreePool CreateEvent SetTimer WaitForEvent SignalEvent CloseEvent CheckEvent InstallProtocolInterface ReinstallProtocolInterface UninstallProtocolInterface HandleProtocol Reserved RegisterProtocolNotify LocateHandle LocateDevicePath InstallConfigurationTable LoadImage StartImage Exit UnloadImage ExitBootServices GetNextMonotonicCount Stall SetWatchdogTimer ConnectController DisconnectController OpenProtocol CloseProtocol OpenProtocolInformation ProtocolsPerHandle LocateHandleBuffer LocateProtocol InstallMultipleProtocolInterfaces UninstallMultipleProtocolInterfaces CalculateCrc32 CopyMem SetMem CreateEventEx')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""Boot Services:
  RaiseTPL                           : 0x{self.RaiseTPL:016X}
  RestoreTPL                         : 0x{self.RestoreTPL:016X}
  AllocatePages                      : 0x{self.AllocatePages:016X}
  FreePages                          : 0x{self.FreePages:016X}
  GetMemoryMap                       : 0x{self.GetMemoryMap:016X}
  AllocatePool                       : 0x{self.AllocatePool:016X}
  FreePool                           : 0x{self.FreePool:016X}
  CreateEvent                        : 0x{self.CreateEvent:016X}
  SetTimer                           : 0x{self.SetTimer:016X}
  WaitForEvent                       : 0x{self.WaitForEvent:016X}
  SignalEvent                        : 0x{self.SignalEvent:016X}
  CloseEvent                         : 0x{self.CloseEvent:016X}
  CheckEvent                         : 0x{self.CheckEvent:016X}
  InstallProtocolInterface           : 0x{self.InstallProtocolInterface:016X}
  ReinstallProtocolInterface         : 0x{self.ReinstallProtocolInterface:016X}
  UninstallProtocolInterface         : 0x{self.UninstallProtocolInterface:016X}
  HandleProtocol                     : 0x{self.HandleProtocol:016X}
  Reserved                           : 0x{self.Reserved:016X}
  RegisterProtocolNotify             : 0x{self.RegisterProtocolNotify:016X}
  LocateHandle                       : 0x{self.LocateHandle:016X}
  LocateDevicePath                   : 0x{self.LocateDevicePath:016X}
  InstallConfigurationTable          : 0x{self.InstallConfigurationTable:016X}
  LoadImage                          : 0x{self.LoadImage:016X}
  StartImage                         : 0x{self.StartImage:016X}
  Exit                               : 0x{self.Exit:016X}
  UnloadImage                        : 0x{self.UnloadImage:016X}
  ExitBootServices                   : 0x{self.ExitBootServices:016X}
  GetNextMonotonicCount              : 0x{self.GetNextMonotonicCount:016X}
  Stall                              : 0x{self.Stall:016X}
  SetWatchdogTimer                   : 0x{self.SetWatchdogTimer:016X}
  ConnectController                  : 0x{self.ConnectController:016X}
  DisconnectController               : 0x{self.DisconnectController:016X}
  OpenProtocol                       : 0x{self.OpenProtocol:016X}
  CloseProtocol                      : 0x{self.CloseProtocol:016X}
  OpenProtocolInformation            : 0x{self.OpenProtocolInformation:016X}
  ProtocolsPerHandle                 : 0x{self.ProtocolsPerHandle:016X}
  LocateHandleBuffer                 : 0x{self.LocateHandleBuffer:016X}
  LocateProtocol                     : 0x{self.LocateProtocol:016X}
  InstallMultipleProtocolInterfaces  : 0x{self.InstallMultipleProtocolInterfaces:016X}
  UninstallMultipleProtocolInterfaces: 0x{self.UninstallMultipleProtocolInterfaces:016X}
  CalculateCrc32                     : 0x{self.CalculateCrc32:016X}
  CopyMem                            : 0x{self.CopyMem:016X}
  SetMem                             : 0x{self.SetMem:016X}
  CreateEventEx                      : 0x{self.CreateEventEx:016X}
"""


# #################################################################################################
# EFI System Configuration Table
# #################################################################################################
#
# \MdePkg\Include\Uefi\UefiSpec.h
# -------------------------------

EFI_VENDOR_TABLE_FORMAT = '<' + EFI_GUID_FMT + 'Q'
EFI_VENDOR_TABLE_SIZE = struct.calcsize(EFI_VENDOR_TABLE_FORMAT)


class EFI_VENDOR_TABLE(namedtuple('EFI_VENDOR_TABLE', 'VendorGuidData VendorTable')):
    __slots__ = ()

    def VendorGuid(self) -> str:
        return EFI_GUID_STR(self.VendorGuidData)


class EFI_CONFIGURATION_TABLE:
    def __init__(self):
        self.VendorTables = {}

    def __str__(self) -> str:
        vendor_table_str = ''.join([f'{{{vt}}} : 0x{self.VendorTables[vt]:016X}\n' for vt in self.VendorTables])
        return f'Vendor Tables:\n{vendor_table_str}'


# #################################################################################################
# EFI DXE Services Table
# #################################################################################################
#
# \MdePkg\Include\Pi\PiDxeCis.h
# -----------------------------
#
EFI_DXE_SERVICES_TABLE_SIGNATURE = 'DXE_SERV'  # 0x565245535f455844
EFI_DXE_SERVICES_TABLE_FMT = '=17Q'


class EFI_DXE_SERVICES_TABLE(namedtuple('EFI_DXE_SERVICES_TABLE', 'AddMemorySpace AllocateMemorySpace FreeMemorySpace RemoveMemorySpace GetMemorySpaceDescriptor SetMemorySpaceAttributes GetMemorySpaceMap AddIoSpace AllocateIoSpace FreeIoSpace RemoveIoSpace GetIoSpaceDescriptor GetIoSpaceMap Dispatch Schedule Trust ProcessFirmwareVolume')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""DXE Services:
  AddMemorySpace          : 0x{self.AddMemorySpace:016X}
  AllocateMemorySpace     : 0x{self.AllocateMemorySpace:016X}
  FreeMemorySpace         : 0x{self.FreeMemorySpace:016X}
  RemoveMemorySpace       : 0x{self.RemoveMemorySpace:016X}
  GetMemorySpaceDescriptor: 0x{self.GetMemorySpaceDescriptor:016X}
  SetMemorySpaceAttributes: 0x{self.SetMemorySpaceAttributes:016X}
  GetMemorySpaceMap       : 0x{self.GetMemorySpaceMap:016X}
  AddIoSpace              : 0x{self.AddIoSpace:016X}
  AllocateIoSpace         : 0x{self.AllocateIoSpace:016X}
  FreeIoSpace             : 0x{self.FreeIoSpace:016X}
  RemoveIoSpace           : 0x{self.RemoveIoSpace:016X}
  GetIoSpaceDescriptor    : 0x{self.GetIoSpaceDescriptor:016X}
  GetIoSpaceMap           : 0x{self.GetIoSpaceMap:016X}
  Dispatch                : 0x{self.Dispatch:016X}
  Schedule                : 0x{self.Schedule:016X}
  Trust                   : 0x{self.Trust:016X}
  ProcessFirmwareVolume   : 0x{self.ProcessFirmwareVolume:016X}
"""


# #################################################################################################
# EFI PEI Services Table
# #################################################################################################
EFI_FRAMEWORK_PEI_SERVICES_TABLE_SIGNATURE = 0x5652455320494550
FRAMEWORK_PEI_SPECIFICATION_MAJOR_REVISION = 0
FRAMEWORK_PEI_SPECIFICATION_MINOR_REVISION = 91
FRAMEWORK_PEI_SERVICES_REVISION = ((FRAMEWORK_PEI_SPECIFICATION_MAJOR_REVISION << 16) | (FRAMEWORK_PEI_SPECIFICATION_MINOR_REVISION))

# #################################################################################################
# EFI System Management System Table
# #################################################################################################

EFI_SMM_SYSTEM_TABLE_SIGNATURE = 'SMST'
EFI_SMM_SYSTEM_TABLE_REVISION = (0 << 16) | (0x09)


EFI_TABLES: Dict[str, Dict[str, Any]] = {
    EFI_SYSTEM_TABLE_SIGNATURE: {'name': 'EFI System Table', 'struct': EFI_SYSTEM_TABLE, 'fmt': EFI_SYSTEM_TABLE_FMT},
    EFI_RUNTIME_SERVICES_SIGNATURE: {'name': 'EFI Runtime Services Table', 'struct': EFI_RUNTIME_SERVICES_TABLE, 'fmt': EFI_RUNTIME_SERVICES_TABLE_FMT},
    EFI_BOOT_SERVICES_SIGNATURE: {'name': 'EFI Boot Services Table', 'struct': EFI_BOOT_SERVICES_TABLE, 'fmt': EFI_BOOT_SERVICES_TABLE_FMT},
    EFI_DXE_SERVICES_TABLE_SIGNATURE: {'name': 'EFI DXE Services Table', 'struct': EFI_DXE_SERVICES_TABLE, 'fmt': EFI_DXE_SERVICES_TABLE_FMT}
    # EFI_FRAMEWORK_PEI_SERVICES_TABLE_SIGNATURE : {'name' : 'EFI Framework PEI Services Table', 'struct' : EFI_FRAMEWORK_PEI_SERVICES_TABLE, 'fmt' : EFI_FRAMEWORK_PEI_SERVICES_TABLE_FMT },
    # EFI_SMM_SYSTEM_TABLE_SIGNATURE             : {'name' : 'EFI SMM System Table',             'struct' : EFI_SMM_SYSTEM_TABLE,             'fmt' : EFI_SMM_SYSTEM_TABLE_FMT             },
    # EFI_CONFIG_TABLE_SIGNATURE                 : {'name' : 'EFI Configuration Table',          'struct' : EFI_CONFIG_TABLE,                 'fmt' : EFI_CONFIG_TABLE_FMT                 }
}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2019-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

__version__ = '1.0'

from typing import List, Optional
from chipsec.library.defines import bit, is_set
from chipsec.hal.hal_base import HALBase
from chipsec.library.exceptions import CSReadError, HWAccessViolationError


class LockResult:
    DEFINED = bit(0)  # lock exists within configuration
    HAS_CONFIG = bit(1)  # lock configuration exists
    LOCKED = bit(2)  # lock matches value within xml
    CAN_READ = bit(3)  # system is able to access the lock
    INCONSISTENT = bit(4)  # all lock results do not match


class locks(HALBase):
    def __init__(self, cs):
        super(locks, self).__init__(cs)

    def get_locks(self) -> List[str]:
        """
        Return a list of locks defined within the configuration file
        """
        return self.cs.lock.get_list()

    def lock_valid(self, lock_name: str, bus: Optional[int] = None) -> int:
        res = 0
        if self.cs.lock.is_defined(lock_name):
            res |= LockResult.DEFINED
        try:
            self.cs.get_locked_value(lock_name)
            self.cs.lock.get(lock_name, bus=bus)
            res |= LockResult.HAS_CONFIG
            res |= LockResult.CAN_READ
        except KeyError:
            pass
        except CSReadError:
            res |= LockResult.HAS_CONFIG
        except HWAccessViolationError:
            res |= LockResult.HAS_CONFIG
        return res

    def is_locked(self, lock_name: str, bus: Optional[int] = None) -> int:
        """
        Return whether the lock has the value setting
        """
        res = self.lock_valid(lock_name, bus)
        if is_set(res, LockResult.HAS_CONFIG) and is_set(res, LockResult.CAN_READ):
            locked = self.cs.get_locked_value(lock_name)
            lock_setting = self.cs.lock.get(lock_name, bus=bus)
            if not all(lock_setting[0] == elem for elem in lock_setting):
                res |= LockResult.INCONSISTENT
            if all(locked == elem for elem in lock_setting):
                res |= LockResult.LOCKED
        return res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2018 - 2021, Intel Corporation
# Copyright (c) 2024, Amd Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
# AMD:
# chipsec@amd.com
#

"""
Displays CPU information

Reference:
    - Intel 64 and IA-32 Architectures Software Developer Manual (SDM)
        - https://www.intel.com/content/www/us/en/developer/articles/technical/intel-sdm.html

Usage:
    ``chipsec_main -m common.cpu.cpu_info``

Examples:
    >>> chipsec_main.py -m common.cpu.cpu_info

Registers used:
    - IA32_BIOS_SIGN_ID.Microcode

.. note:
    No PASS/FAIL returned, INFORMATION only.

"""

import struct
from chipsec.module_common import BaseModule, CPU
from chipsec.library.returncode import ModuleResult
from chipsec.library.defines import bytestostring
from typing import List

TAGS = [CPU]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'COMMON', 'CPU', 'CPU_INFO']


class cpu_info(BaseModule):
    def __init__(self):
        super(cpu_info, self).__init__()
        self.result.url = 'https://chipsec.github.io/modules/chipsec.modules.common.cpu.cpu_info.html'

    def is_supported(self) -> bool:
        if self.cs.is_intel():
            if self.cs.register.has_field('IA32_BIOS_SIGN_ID', 'Microcode'):
                return True
            self.logger.log_important('IA32_BIOS_SIGN_ID.Microcode not defined for platform.  Skipping module.')
        else:
            return self.cs.is_amd()
        return False

    def run_intel(self, module_argv: List[str]) -> int:
        thread_count = 1
        if not self.cs.os_helper.is_efi():
            thread_count = self.cs.msr.get_cpu_thread_count()

        for thread in range(thread_count):
            # Handle processor binding so we are always checking processor 0
            # for this example.  No need to do this in UEFI Shell.
            if not self.cs.os_helper.is_efi():
                self.cs.helper.set_affinity(thread)

            self.logger.log(f'[*] Thread {thread:04d}')

            brand = ''
            for eax_val in [0x80000002, 0x80000003, 0x80000004]:
                regs = self.cs.cpu.cpuid(eax_val, 0)
                for i in range(4):
                    brand += bytestostring(struct.pack('<I', regs[i]))
            brand = brand.rstrip('\x00')
            self.logger.log(f'[*] Processor: {brand}')

            (eax, _, _, _) = self.cs.cpu.cpuid(0x01, 0x00)
            stepping = eax & 0xF
            model = (eax >> 4) & 0xF
            family = (eax >> 8) & 0xF
            if (family == 0x0F) or (family == 0x06):
                model = ((eax >> 12) & 0xF0) | model
            if family == 0x0F:
                family = ((eax >> 20) & 0xFF) | family
            self.logger.log(f'[*]            Family: {family:02X} Model: {model:02X} Stepping: {stepping:01X}')

            microcode_rev = self.cs.register.read_field('IA32_BIOS_SIGN_ID', 'Microcode', cpu_thread=thread)
            self.logger.log(f'[*]            Microcode: {microcode_rev:08X}')
            self.logger.log('[*]')

        self.logger.log_information('Processor information displayed')

        self.result.setStatusBit(self.result.status.INFORMATION)
        return self.result.getReturnCode(ModuleResult.INFORMATION)

    def run_amd(self) -> int:
        # Determine number of threads to check
        thread_count = 1
        if not self.cs.os_helper.is_efi():
            (_, _, r_rcx, _) = self.cs.cpu.cpuid(0x80000008, 0)
            thread_count = r_rcx & 0xff

        for thread in range(thread_count):
            # Handle processor binding so we are always checking processor 0
            # for this example.  No need to do this in UEFI Shell.
            if not self.cs.os_helper.is_efi():
                self.cs.helper.set_affinity(thread)

            # Display thread
            self.logger.log(f'[*] Thread {thread:04d}')

            # Get processor brand string
            brand = ''
            brand = brand.rstrip('\x00')
            self.logger.log(f'[*] Processor: {brand}')

            # "Authentic AMD"
            e_rbx = int("htuA".encode('utf-8').hex(), 16)    #0x68747541
            e_rcx = int("DMAc".encode('utf-8').hex(), 16)    #0x444D4163
            e_rdx = int("itne".encode('utf-8').hex(), 16)    #0x69746E65
            (_, r_rbx, r_rcx, r_rdx) = self.cs.cpu.cpuid(0x00, 0x00)

            if not (e_rbx == r_rbx and e_rcx == r_rcx and e_rdx == r_rdx):
                self.logger.log_failed("Not Authentic AMD")
                self.res = ModuleResult.FAILED
                return self.res

            # Get processor version information
            (r_rax, _, _, _) = self.cs.cpu.cpuid(0x01, 0x00)
            baseModel = (r_rax >> 4) & 0xF
            baseFamily = (r_rax >> 8) & 0xF
            extModel = (r_rax >> 16) & 0xF
            extFamily = (r_rax >> 20) & 0xF
            stepping = (r_rax) & 0xF
            family = baseFamily + extFamily
            model = (extModel << 4) & baseModel
            self.logger.log(f'[*]            Family: {family:02X} Model: {model:02X} Stepping: {stepping:01X}')

            # Get microcode revision
            microcode_rev = self.cs.register.read_field('PATCH_LEVEL', 'PatchLevel', cpu_thread=thread)
            self.logger.log(f'[*]            Microcode: {microcode_rev:08X}')
            self.logger.log('[*]')
        self.logger.log_information('Processor information displayed')

        self.res = ModuleResult.INFORMATION
        return self.res

    def run(self, module_argv: List[str]) -> int:
        # Log the start of the test
        self.logger.start_test('Current Processor Information:')

        if self.cs.is_intel():
            return self.run_intel(module_argv)
        else:
            return self.run_amd()



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Access to virtual memory

usage:
    >>> read_virtual_mem( 0xf0000, 0x100 )
    >>> write_virtual_mem( 0xf0000, 0x100, buffer )
    >>> write_virtual_mem_dowrd( 0xf0000, 0xdeadbeef )
    >>> read_virtual_mem_dowrd( 0xfed40000 )
"""

import struct
from typing import Tuple
from chipsec.library.logger import logger, print_buffer_bytes
from chipsec.hal import hal_base


class VirtMemory(hal_base.HALBase):
    def __init__(self, cs):
        super(VirtMemory, self).__init__(cs)
        self.helper = cs.helper

    ####################################################################################
    #
    # virtual memory API using 64b virtual Address
    # (Same functions as below just using 64b PA instead of High and Low 32b parts of PA)
    #
    ####################################################################################

    # Reading virtual memory

    def read_virtual_mem(self, virt_address: int, length: int) -> int:
        logger().log_hal(f'[mem] 0x{virt_address:016X}')
        phys_address = self.va2pa(virt_address)
        return self.helper.read_phys_mem(phys_address, length)

    def read_virtual_mem_dword(self, virt_address: int) -> int:
        phys_address = self.va2pa(virt_address)
        out_buf = self.helper.read_phys_mem(phys_address, 4)
        value = struct.unpack('=I', out_buf)[0]
        logger().log_hal(f'[mem] dword at VA = 0x{virt_address:016X}: 0x{value:08X}')
        return value

    def read_virtual_mem_word(self, virt_address: int) -> int:
        phys_address = self.va2pa(virt_address)
        out_buf = self.helper.read_phys_mem(phys_address, 2)
        value = struct.unpack('=H', out_buf)[0]
        logger().log_hal(f'[mem] word at VA = 0x{virt_address:016X}: 0x{value:04X}')
        return value

    def read_virtual_mem_byte(self, virt_address: int) -> int:
        phys_address = self.va2pa(virt_address)
        out_buf = self.helper.read_phys_mem(phys_address, 1)
        value = struct.unpack('=B', out_buf)[0]
        logger().log_hal(f'[mem] byte at VA = 0x{virt_address:016X}: 0x{value:02X}')
        return value

    # Writing virtual memory

    def write_virtual_mem(self, virt_address: int, length: int, buf: bytes) -> int:
        logger().log_hal(f'[mem] buffer len = 0x{length:X} to VA = 0x{virt_address:016X}')
        if logger().HAL:
            print_buffer_bytes(buf)
        phys_address = self.va2pa(virt_address)
        return self.helper.write_phys_mem(phys_address, length, buf)

    def write_virtual_mem_dword(self, virt_address: int, dword_value: int) -> int:
        logger().log_hal(f'[mem] dword to VA = 0x{virt_address:016X} <- 0x{dword_value:08X}')
        phys_address = self.va2pa(virt_address)
        return self.helper.write_phys_mem(phys_address, 4, struct.pack('I', dword_value))

    def write_virtual_mem_word(self, virt_address: int, word_value: int) -> int:
        logger().log_hal(f'[mem] word to VA = 0x{virt_address:016X} <- 0x{word_value:04X}')
        phys_address = self.va2pa(virt_address)
        return self.helper.write_phys_mem(phys_address, 2, struct.pack('H', word_value))

    def write_virtual_mem_byte(self, virt_address: int, byte_value: int) -> int:
        logger().log_hal(f'[mem] byte to VA = 0x{virt_address:016X} <- 0x{byte_value:02X}')
        phys_address = self.va2pa(virt_address)
        return self.helper.write_phys_mem(phys_address, 1, struct.pack('B', byte_value))

    # Allocate virtual memory buffer

    def alloc_virtual_mem(self, length: int, max_phys_address: int = 0xFFFFFFFFFFFFFFFF) -> Tuple[int, int]:
        (va, pa) = self.helper.alloc_phys_mem(length, max_phys_address)
        logger().log_hal(f'[mem] Allocated: PA = 0x{pa:016X}, VA = 0x{va:016X}')
        return (va, pa)

    def va2pa(self, va: int) -> int:
        (pa, error_code) = self.helper.va2pa(va)
        if error_code:
            logger().log_hal(f'[mem] Looks like VA (0x{va:016X}) not mapped')
            return va
        logger().log_hal(f'[mem] VA (0x{va:016X}) -> PA (0x{pa:016X})')
        return pa

    def free_virtual_mem(self, virt_address: int) -> bool:
        pa = self.va2pa(virt_address)
        ret = self.helper.free_phys_mem(pa)
        logger().log_hal(f'[mem] Deallocated : VA = 0x{virt_address:016X}')
        return ret == 1



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

import os
import configparser
from typing import Any, List
from chipsec.library.file import get_main_dir
from chipsec.library.exceptions import CSConfigError


class NoDefault():
    pass


class Options(object):

    def __init__(self):
        options_path = os.path.join(get_main_dir(), 'chipsec', 'options')
        if not os.path.isdir(options_path):
            raise CSConfigError(f'Unable to locate configuration options: {options_path}')
        options_name = os.path.join(options_path, 'cmd_options.ini')
        self.config = configparser.ConfigParser()
        with open(options_name) as options_file:
            self.config.read_file(options_file)

    def get_section_data(self, section: str, key: str, default: Any = NoDefault) -> str:
        try:
            ret_data = self.config.get(section, key)
        except Exception as e:
            if default is NoDefault:
                raise e
            return default
        return ret_data

    def get_list_data(self, list_section: str, list_key: str, list_default: Any, list_separator: str = ',') -> List[Any]:
        raw_data = self.get_section_data(list_section, list_key, '').split(list_separator)
        data = [item.strip() for item in raw_data]
        return data if data[0] else list_default



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2019, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Tests that IA-32/IA-64 architectural features are configured and locked, including IA32 Model Specific Registers (MSRs)

Reference:
    - Intel 64 and IA-32 Architectures Software Developer Manual (SDM)
        - https://www.intel.com/content/www/us/en/developer/articles/technical/intel-sdm.html

Usage:
    ``chipsec_main -m common.ia32cfg``

Examples:
    >>> chipsec_main.py -m common.ia32cfg

Registers used:
    - IA32_FEATURE_CONTROL
    - Ia32FeatureControlLock (control)

"""

from chipsec.module_common import BaseModule, HWCONFIG
from chipsec.library.returncode import ModuleResult
from typing import List

TAGS = [HWCONFIG]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'COMMON', 'IA32CFG']


class ia32cfg(BaseModule):
    def __init__(self):
        BaseModule.__init__(self)
        self.result.url = 'https://chipsec.github.io/modules/chipsec.modules.common.ia32cfg.html'
        self.res = ModuleResult.PASSED

    def is_supported(self) -> bool:
        if self.cs.register.is_defined('IA32_FEATURE_CONTROL'):
            if self.cs.control.is_defined('Ia32FeatureControlLock'):
                return True
            self.logger.log_important('Ia32FeatureControlLock control not defined for platform.  Skipping module.')
        else:
            self.logger.log_important('IA32_FEATURE_CONTROL register not defined for platform.  Skipping module.')
        return False

    def check_ia32feature_control(self) -> int:
        self.logger.log('[*] Verifying IA32_Feature_Control MSR is locked on all logical CPUs..')

        res = ModuleResult.PASSED
        for tid in range(self.cs.msr.get_cpu_thread_count()):
            if self.logger.VERBOSE:
                feature_cntl = self.cs.register.read('IA32_FEATURE_CONTROL', tid)
                self.cs.register.print('IA32_FEATURE_CONTROL', feature_cntl)
            feature_cntl_lock = self.cs.control.get('Ia32FeatureControlLock', tid)
            self.logger.log(f'[*] cpu{tid:d}: IA32_FEATURE_CONTROL Lock = {feature_cntl_lock:d}')
            if 0 == feature_cntl_lock:
                res = ModuleResult.FAILED
                self.result.setStatusBit(self.result.status.LOCKS)

        if res == ModuleResult.PASSED:
            self.logger.log_passed('IA32_FEATURE_CONTROL MSR is locked on all logical CPUs')
        else:
            self.logger.log_failed('IA32_FEATURE_CONTROL MSR is not locked on all logical CPUs')

        return self.result.getReturnCode(res)

    def run(self, module_argv: List[str]) -> int:
        self.logger.start_test('IA32 Feature Control Lock')
        self.res = self.check_ia32feature_control()
        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2020, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Pretty simple VMM hypercall fuzzer

Usage:
    ``chipsec_main.py -i -m tools.vmm.hypercallfuzz [-a <mode>,<vector_reg>,<maxval>,<iterations>]``

    - ``mode``           : Hypercall fuzzing mode
        * ``exhaustive`` : Fuzz all arguments exhaustively in range ``[0:<maxval>]`` (default)
        * ``random``     : Send random values in all registers in range ``[0:<maxval>]``
    - ``vector_reg``     : Hypercall vector register
    - ``maxval``         : Maximum value of each register
    - ``iterations``     : Number of iterations in random mode

Where:
    - ``[]``: optional line

Examples:
    >>> chipsec_main.py -i -m tools.vmm.hypercallfuzz
    >>> chipsec_main.py -i -m tools.vmm.hypercallfuzz -a random,22,0xFFFF,1000

Additional options set within the module:
    - ``DEFAULT_VECTOR_MAXVAL``     : Default maximum value
    - ``DEFAULT_MAXVAL_EXHAUSTIVE`` : Default maximum value for exhaustive testing
    - ``DEFAULT_MAXVAL_RANDOM``     : Default maximum value for random testing
    - ``DEFAULT_RANDOM_ITERATIONS`` : Default iterations for random testing
    - ``_FLUSH_LOG_EACH_ITER``      : Set to flush log after each iteration
    - ``_LOG_ALL_GPRS``             : Display log of each iteration values

.. note::
    - Returns a Warning by default
    - System may be in an unknown state, further evaluation may be needed

.. important::
    - This module is designed to run in a VM environment
    - Behavior on physical HW is undefined

"""

import random
import time

from chipsec.module_common import BaseModule
from chipsec.library.returncode import ModuleResult
from chipsec.hal.vmm import VMM

DEFAULT_VECTOR_MAXVAL = 0xFF

DEFAULT_MAXVAL_EXHAUSTIVE = 0xFF
DEFAULT_MAXVAL_RANDOM = 0xFFFFFFFF

DEFAULT_RANDOM_ITERATIONS = 0x7FFFFFFF

# Flush log file before each port
_FLUSH_LOG_EACH_ITER = False
_LOG_ALL_GPRS = True

GPRS = {'rax': 0, 'rbx': 0, 'rcx': 0, 'rdx': 0, 'rdi': 0, 'rsi': 0, 'r8': 0, 'r9': 0, 'r10': 0, 'r11': 0}


class hypercallfuzz(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)
        self.vmm = VMM(self.cs)
        self.random_order = True
        self.gprs = GPRS
        self.vector_reg = None
        self.iterations = DEFAULT_RANDOM_ITERATIONS
        self.maxval = DEFAULT_MAXVAL_RANDOM

    def is_supported(self):
        return True

    def fuzz_generic_hypercalls(self):
        _fmt = '{:02X}' if self.maxval <= 0xFF else ('{:04X}' if self.maxval <= 0xFFFF else ('{:08X}' if self.maxval <= 0xFFFFFFFF else '{:016X}'))
        _str = f'{{:d}} hcall rax={_fmt},rbx={_fmt},rcx={_fmt},rdx={_fmt},rdi={_fmt},rsi={_fmt},r8={_fmt},r9={_fmt},r10={_fmt},r11={_fmt}'

        t = time.time()
        if self.random_order:

            self.logger.log(f'[*] Fuzzing {self.iterations:d} random hypercalls with random arguments...')
            for it in range(self.iterations):
                rax = random.randint(0, self.gprs['rax'])
                rbx = random.randint(0, self.gprs['rbx'])
                rcx = random.randint(0, self.gprs['rcx'])
                rdx = random.randint(0, self.gprs['rdx'])
                rdi = random.randint(0, self.gprs['rdi'])
                rsi = random.randint(0, self.gprs['rsi'])
                r8 = random.randint(0, self.gprs['r8'])
                r9 = random.randint(0, self.gprs['r9'])
                r10 = random.randint(0, self.gprs['r10'])
                r11 = random.randint(0, self.gprs['r11'])
                if _LOG_ALL_GPRS:
                    self.logger.log(_str.format(it, rax, rbx, rcx, rdx, rdi, rsi, r8, r9, r10, r11))
                else:
                    self.logger.log(f'{it:d} hcall')
                if _FLUSH_LOG_EACH_ITER:
                    self.logger.flush()
                try:
                    self.vmm.hypercall(rax, rbx, rcx, rdx, rdi, rsi, r8, r9, r10, r11)
                except:
                    pass
        else:
            it = 0
            self.logger.log('[*] Fuzzing hypercalls with arguments exhaustively...')
            for rax in range(self.gprs['rax']):
                for rbx in range(self.gprs['rbx']):
                    for rcx in range(self.gprs['rcx']):
                        for rdx in range(self.gprs['rdx']):
                            for rdi in range(self.gprs['rdi']):
                                for rsi in range(self.gprs['rsi']):
                                    for r8 in range(self.gprs['r8']):
                                        for r9 in range(self.gprs['r9']):
                                            for r10 in range(self.gprs['r10']):
                                                for r11 in range(self.gprs['r11']):
                                                    if _LOG_ALL_GPRS:
                                                        self.logger.log(_str.format(it, rax, rbx, rcx, rdx, rdi, rsi, r8, r9, r10, r11))
                                                    else:
                                                        self.logger.log(f'{it:d} hcall')
                                                    if _FLUSH_LOG_EACH_ITER:
                                                        self.logger.flush()
                                                    try:
                                                        self.vmm.hypercall(rax, rbx, rcx, rdx, rdi, rsi, r8, r9, r10, r11)
                                                        it += 1
                                                    except:
                                                        pass

        self.logger.log(f'[*] Finished fuzzing: time elapsed {time.time() - t:.3f}')

    def run(self, module_argv):
        self.logger.start_test('Dumb VMM hypercall fuzzer')

        if len(module_argv) > 0:
            self.random_order = module_argv[0].lower() == 'random'
        self.maxval = DEFAULT_MAXVAL_RANDOM if self.random_order else DEFAULT_MAXVAL_EXHAUSTIVE
        if len(module_argv) > 1:
            self.vector_reg = module_argv[1]
        if len(module_argv) > 2:
            self.maxval = int(module_argv[2], 16)
        if len(module_argv) > 3:
            self.iterations = int(module_argv[3])

        for r in self.gprs:
            self.gprs[r] = self.maxval
        if self.vector_reg is not None:
            self.gprs[self.vector_reg] = DEFAULT_VECTOR_MAXVAL

        self.logger.log('\n[*] Configuration:')
        self.logger.log(f'    Mode               : {"random" if self.random_order else "exhaustive"}')
        self.logger.log(f'    Hypercall vector in: {self.vector_reg}')
        self.logger.log(f'    Max register value : 0x{self.maxval:X}')
        self.logger.log(f'    Iterations         : {self.iterations:d}\n')

        self.res = self.fuzz_generic_hypercalls()

        self.logger.log_information('Module completed')
        self.logger.log_warning('System may be in an unknown state, further evaluation may be needed.')
        self.result.setStatusBit(self.result.status.VERIFY)
        return self.result.getReturnCode(ModuleResult.WARNING)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
SMI command:

>>> chipsec_util smi count
>>> chipsec_util smi send <thread_id> <SMI_code> <SMI_data> [RAX] [RBX] [RCX] [RDX] [RSI] [RDI]
>>> chipsec_util smi smmc <RT_code_start> <RT_code_end> <GUID> <payload_loc> <payload_file|payload_string> [port]

Examples:

>>> chipsec_util smi count
>>> chipsec_util smi send 0x0 0xDE 0x0
>>> chipsec_util smi send 0x0 0xDE 0x0 0xAAAAAAAAAAAAAAAA ..
>>> chipsec_util smi smmc 0x79dfe000 0x79efdfff ed32d533-99e6-4209-9cc02d72cdd998a7 0x79dfaaaa payload.bin

NMI command:

>>> chipsec_util nmi

Examples:

>>> chipsec_util nmi
"""

import os

from chipsec.command import BaseCommand, toLoad
from chipsec.hal.interrupts import Interrupts
from chipsec.hal.uefi_common import EFI_ERROR_STR
from argparse import ArgumentParser


# ###################################################################
#
# CPU Interrupts
#
# ###################################################################


class SMICommand(BaseCommand):
    """
    >>> chipsec_util smi count
    >>> chipsec_util smi send <thread_id> <SMI_code> <SMI_data> [RAX] [RBX] [RCX] [RDX] [RSI] [RDI]
    >>> chipsec_util smi smmc <RT_code_start> <RT_code_end> <GUID> <payload_loc> <payload_file|payload_string> [port]

    Examples:

    >>> chipsec_util smi count
    >>> chipsec_util smi send 0x0 0xDE 0x0
    >>> chipsec_util smi send 0x0 0xDE 0x0 0xAAAAAAAAAAAAAAAA ..
    >>> chipsec_util smi smmc 0x79dfe000 0x79efdfff ed32d533-99e6-4209-9cc02d72cdd998a7 0x79dfaaaa payload.bin
    """

    def requirements(self) -> toLoad:
        return toLoad.All

    def parse_arguments(self) -> None:
        parser = ArgumentParser(prog='chipsec_util smi', usage=SMICommand.__doc__)
        subparsers = parser.add_subparsers()

        parser_count = subparsers.add_parser('count')
        parser_count.set_defaults(func=self.smi_count)

        parser_send = subparsers.add_parser('send')
        parser_send.add_argument('thread_id', type=lambda x: int(x, 16), help='Thread ID (hex)')
        parser_send.add_argument('SMI_code_port_value', type=lambda x: int(x, 16), help='SMI Code (hex)')
        parser_send.add_argument('SMI_data_port_value', type=lambda x: int(x, 16), help='SMI Data (hex)')
        parser_send.add_argument('_rax', type=lambda x: int(x, 16), nargs='?', default=None, help='RAX (hex)')
        parser_send.add_argument('_rbx', type=lambda x: int(x, 16), nargs='?', default=0, help='RBX (hex) [default=0]')
        parser_send.add_argument('_rcx', type=lambda x: int(x, 16), nargs='?', default=0, help='RCX (hex) [default=0]')
        parser_send.add_argument('_rdx', type=lambda x: int(x, 16), nargs='?', default=0, help='RDX (hex) [default=0]')
        parser_send.add_argument('_rsi', type=lambda x: int(x, 16), nargs='?', default=0, help='RSI (hex) [default=0]')
        parser_send.add_argument('_rdi', type=lambda x: int(x, 16), nargs='?', default=0, help='RDI (hex) [default=0]')
        parser_send.set_defaults(func=self.smi_send)

        parser_smmc = subparsers.add_parser('smmc')
        parser_smmc.add_argument('RTC_start', type=lambda x: int(x, 16), help='RTC Code Start (hex)')
        parser_smmc.add_argument('RTC_end', type=lambda x: int(x, 16), help='RT Code End (hex)')
        parser_smmc.add_argument('guid', type=str, help='GUID')
        parser_smmc.add_argument('payload_loc', type=lambda x: int(x, 16), help='Payload Location (hex)')
        parser_smmc.add_argument('payload', type=str, help='Payload')
        parser_smmc.add_argument('port', type=lambda x: int(x, 16), nargs='?', default=0x0, help='Port (hex) [default=0]')
        parser_smmc.set_defaults(func=self.smi_smmc)

        parser.parse_args(self.argv, namespace=self)

    def smi_count(self) -> None:
        self.logger.log("[CHIPSEC] SMI count:")
        for tid in range(self.cs.msr.get_cpu_thread_count()):
            smi_cnt = self.cs.register.read_field('MSR_SMI_COUNT', 'Count', cpu_thread=tid)
            self.logger.log(f'  CPU{tid:d}: {smi_cnt:d}')

    def smi_smmc(self) -> None:
        if os.path.isfile(self.payload):
            with open(self.payload, 'rb') as f:
                self.payload = f.read()

        self.logger.log(f'Searching for \'smmc\' in range 0x{self.RTC_start:x}-0x{self.RTC_end:x}')
        # scan for SMM_CORE_PRIVATE_DATA smmc signature
        smmc_loc = self.interrupts.find_smmc(self.RTC_start, self.RTC_end)
        if (smmc_loc == 0):
            self.logger.log(" Couldn't find smmc signature")
            return
        self.logger.log(f'Found \'smmc\' structure at 0x{smmc_loc:x}')

        ReturnStatus = self.interrupts.send_smmc_SMI(smmc_loc, self.guid, self.payload, self.payload_loc, CommandPort=self.port)
        # TODO Translate ReturnStatus to EFI_STATUS enum
        self.logger.log(f'ReturnStatus: 0x{ReturnStatus:x} ({EFI_ERROR_STR(ReturnStatus)})')

    def smi_send(self) -> None:
        self.logger.log(f'[CHIPSEC] Sending SW SMI (code: 0x{self.SMI_code_port_value:02X}, data: 0x{self.SMI_data_port_value:02X})..')
        if self._rax is None:
            self.interrupts.send_SMI_APMC(self.SMI_code_port_value, self.SMI_data_port_value)
        else:
            self.logger.log(f'          RAX: 0x{self._rax:016X} (AX will be overridden with values of SW SMI ports B2/B3)')
            self.logger.log(f'          RBX: 0x{self._rbx:016X}')
            self.logger.log(f'          RCX: 0x{self._rcx:016X}')
            self.logger.log(f'          RDX: 0x{self._rdx:016X} (DX will be overridden with 0x00B2)')
            self.logger.log(f'          RSI: 0x{self._rsi:016X}')
            self.logger.log(f'          RDI: 0x{self._rdi:016X}')
            ret = self.interrupts.send_SW_SMI(self.thread_id, self.SMI_code_port_value, self.SMI_data_port_value, self._rax, self._rbx, self._rcx, self._rdx, self._rsi, self._rdi)
            if not ret is None:
                self.logger.log("Return values")
                self.logger.log(f'          RAX: {ret[1]:16X}')
                self.logger.log(f'          RBX: {ret[2]:16X}')
                self.logger.log(f'          RCX: {ret[3]:16X}')
                self.logger.log(f'          RDX: {ret[4]:16X}')
                self.logger.log(f'          RSI: {ret[5]:16X}')
                self.logger.log(f'          RDI: {ret[6]:16X}')

    def run(self) -> None:
        try:
            self.interrupts = Interrupts(self.cs)
        except RuntimeError as msg:
            self.logger.log(msg)
            return

        self.func()


class NMICommand(BaseCommand):
    """
    >>> chipsec_util nmi

    Examples:

    >>> chipsec_util nmi
    """

    def requirements(self) -> toLoad:
        return toLoad.All

    def parse_arguments(self) -> None:
        return

    def run(self) -> None:
        try:
            interrupts = Interrupts(self.cs)
        except RuntimeError as msg:
            self.logger.log(msg)
            return

        self.logger.log("[CHIPSEC] Sending NMI#...")
        interrupts.send_NMI()


commands = {'smi': SMICommand, 'nmi': NMICommand}



from lxml import etree
from typing import Any, Dict, List, Tuple

def parse_testsuite(tree: etree) -> Dict[Any, Any]:
    st = tree.xpath('//testsuites/testsuite')
    assert(len(st) == 1),"more than one test-suite in XML file"
    st = st[0]
    res = {}
    res.update(st.attrib)
    res.update(dict([(x.attrib['name'], x.attrib['value']) for x in st.iterfind('.//properties/property')]))
    return res

def parse_test_cases(tree: etree) -> List[Dict[Any, Any]]:
    st = tree.xpath('//testsuites/testsuite')[0]
    entries = []
    for case in st.iterfind('.//testcase'):
        res = {}
        res.update(case.attrib)
        res['pass'] = case.xpath('.//pass')[0].attrib['type']
        res['out'] = case.xpath('.//system-out')[0].text
        if res['out'] is not None:
            res['out'] = res['out']
        else:
            res['out'] = ""
        entries.append(res)
    return entries


def parse_chipsec_xml(fdlike: str) -> Tuple[Dict[Any, Any], List[Any]]:
    if isinstance(fdlike,str):
        fd = open(fdlike, 'rb')
    else:
        fd = fdlike

    tree = etree.parse(fd)
    suite_data = parse_testsuite(tree)
    cases_data = parse_test_cases(tree)
    return (suite_data, cases_data)





# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
The pci command can enumerate PCI/PCIe devices, enumerate expansion ROMs and allow direct access to PCI configuration registers via bus/device/function.

>>> chipsec_util pci enumerate
>>> chipsec_util pci read <bus> <device> <function> <offset> [width]
>>> chipsec_util pci write <bus> <device> <function> <offset> <width> <value>
>>> chipsec_util pci dump [<bus>] [<device>] [<function>]
>>> chipsec_util pci xrom [<bus>] [<device>] [<function>] [xrom_address]
>>> chipsec_util pci cmd [mask] [class] [subclass]

Examples:

>>> chipsec_util pci enumerate
>>> chipsec_util pci read 0 0 0 0x00
>>> chipsec_util pci read 0 0 0 0x88 byte
>>> chipsec_util pci write 0 0x1F 0 0xDC 1 0x1
>>> chipsec_util pci write 0 0 0 0x98 dword 0x004E0040
>>> chipsec_util pci dump
>>> chipsec_util pci dump 0 0 0
>>> chipsec_util pci xrom
>>> chipsec_util pci xrom 3 0 0 0xFEDF0000
>>> chipsec_util pci cmd
>>> chipsec_util pci cmd 1
"""

from chipsec.command import BaseCommand, toLoad
from chipsec.library.logger import pretty_print_hex_buffer
from argparse import ArgumentParser
from chipsec_util import get_option_width, is_option_valid_width, CMD_OPTS_WIDTH
from chipsec.hal.pci import print_pci_devices, print_pci_XROMs
from chipsec.hal.pci import PCI_HDR_CLS_OFF, PCI_HDR_SUB_CLS_OFF, PCI_HDR_CMD_OFF

# PCIe Devices and Configuration Registers


class PCICommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.Driver

    def parse_arguments(self) -> None:
        parser = ArgumentParser(prog='chipsec_util pci', usage=__doc__)
        subparsers = parser.add_subparsers()
        parser_enumerate = subparsers.add_parser('enumerate')
        parser_enumerate.set_defaults(func=self.pci_enumerate)

        parser_read = subparsers.add_parser('read')
        parser_read.add_argument('bus', type=lambda x: int(x, 16), help='Bus (hex)')
        parser_read.add_argument('device', type=lambda x: int(x, 16), help='Device (hex)')
        parser_read.add_argument('function', type=lambda x: int(x, 16), help='Function (hex)')
        parser_read.add_argument('offset', type=lambda x: int(x, 16), help='Offset (hex)')
        parser_read.add_argument('size', type=str, default=None, nargs='?', help='Width')
        parser_read.set_defaults(func=self.pci_read)

        parser_write = subparsers.add_parser('write')
        parser_write.add_argument('bus', type=lambda x: int(x, 16), help='Bus (hex)')
        parser_write.add_argument('device', type=lambda x: int(x, 16), help='Device (hex)')
        parser_write.add_argument('function', type=lambda x: int(x, 16), help='Function (hex)')
        parser_write.add_argument('offset', type=lambda x: int(x, 16), help='Offset (hex)')
        parser_write.add_argument('size', type=str, help='Width')
        parser_write.add_argument('value', type=lambda x: int(x, 16), help='Value (hex)')
        parser_write.set_defaults(func=self.pci_write)

        parser_dump = subparsers.add_parser('dump')
        parser_dump.add_argument('bus', type=lambda x: int(x, 16), nargs='?', default=None, help='Bus (hex)')
        parser_dump.add_argument('device', type=lambda x: int(x, 16), nargs='?', default=None, help='Device (hex)')
        parser_dump.add_argument('function', type=lambda x: int(x, 16), nargs='?', default=None, help='Function (hex)')
        parser_dump.set_defaults(func=self.pci_dump)

        parser_xrom = subparsers.add_parser('xrom')
        parser_xrom.add_argument('bus', type=lambda x: int(x, 16), nargs='?', default=None, help='Bus (hex)')
        parser_xrom.add_argument('device', type=lambda x: int(x, 16), nargs='?', default=None, help='Device (hex)')
        parser_xrom.add_argument('function', type=lambda x: int(x, 16), nargs='?', default=None, help='Function (hex)')
        parser_xrom.add_argument('xrom_addr', type=lambda x: int(x, 16), nargs='?', default=None, help='XROM Address (hex)')
        parser_xrom.set_defaults(func=self.pci_xrom)

        parser_cmd = subparsers.add_parser('cmd')
        parser_cmd.add_argument('cmd_mask', type=lambda x: int(x, 16), default=0xFFFF, nargs='?', help='Mask (hex)')
        parser_cmd.add_argument('pci_class', type=lambda x: int(x, 16), default=None, nargs='?', help='Class (hex)')
        parser_cmd.add_argument('pci_sub_class', type=lambda x: int(x, 16), default=None, nargs='?', help='Subclass (hex)')
        parser_cmd.set_defaults(func=self.pci_cmd)

        parser.parse_args(self.argv, namespace=self)

    def pci_enumerate(self):
        self.logger.log("[CHIPSEC] Enumerating available PCIe devices...")
        print_pci_devices(self.cs.pci.enumerate_devices())

    def pci_dump(self):
        if self.bus is not None:
            if self.device is not None and self.function is not None:
                devices = [(self.bus, self.device, self.function, 0x0000, 0x0000, 0x0000)]
            else:
                devices = self.cs.pci.enumerate_devices(self.bus, self.device, self.function)

            for (_bus, _device, _function, _vid, _did, _rid) in devices:
                self.logger.log("[CHIPSEC] PCI device {:02X}:{:02X}.{:02X} configuration:".format(_bus, _device, _function))
                cfg_buf = self.cs.pci.dump_pci_config(_bus, _device, _function)
                pretty_print_hex_buffer(cfg_buf)
        else:
            self.logger.log("[CHIPSEC] Dumping configuration of available PCI devices...")
            self.cs.pci.print_pci_config_all()

    def pci_xrom(self):
        if self.bus is not None:
            if self.device is not None and self.function is not None:
                devices = [(self.bus, self.device, self.function, 0x0000, 0x0000, 0x000)]
            else:
                devices = self.cs.pci.enumerate_devices(self.bus, self.device, self.function)

            for (_bus, _device, _function, _vid, _did, _rid) in devices:
                self.logger.log("[CHIPSEC] Locating PCI expansion ROM (XROM) of {:02X}:{:02X}.{:02X}...".format(_bus, _device, _function))
                exists, xrom = self.cs.pci.find_XROM(_bus, _device, _function, True, True, self.xrom_addr)
                if exists:
                    self.logger.log("[CHIPSEC] Found XROM of {:02X}:{:02X}.{:02X}".format(_bus, _device, _function))
                    if xrom is not None:
                        self.logger.log("[CHIPSEC] XROM enabled = {:d}, base = 0x{:08X}, size = 0x{:08X}".format(xrom.en, xrom.base, xrom.size))
                        if xrom.header is not None:
                            self.logger.log("[CHIPSEC] XROM header: {}".format(xrom.header))
                else:
                    self.logger.log("[CHIPSEC] Couldn't find XROM of {:02X}:{:02X}.{:02X}".format(_bus, _device, _function))
        else:
            self.logger.log("[CHIPSEC] Enumerating PCI expansion ROMs...")
            _xroms = self.cs.pci.enumerate_xroms(True, True, self.xrom_addr)
            self.logger.log("[CHIPSEC] found {:d} PCI expansion ROMs".format(len(_xroms)))
            if len(_xroms) > 0:
                print_pci_XROMs(_xroms)

    def pci_read(self):
        width = 4
        if self.size is not None:
            width = get_option_width(self.size) if is_option_valid_width(self.size) else int(self.size, 16)

        if 1 == width:
            pci_value = self.cs.pci.read_byte(self.bus, self.device, self.function, self.offset)
        elif 2 == width:
            pci_value = self.cs.pci.read_word(self.bus, self.device, self.function, self.offset)
        elif 4 == width:
            pci_value = self.cs.pci.read_dword(self.bus, self.device, self.function, self.offset)
        else:
            self.logger.log_error("Width should be one of {}".format(CMD_OPTS_WIDTH))
            return
        self.logger.log("[CHIPSEC] PCI {:02X}:{:02X}.{:02X} + 0x{:02X}: 0x{:X}".format(self.bus, self.device, self.function, self.offset, pci_value))

    def pci_write(self):
        width = get_option_width(self.size) if is_option_valid_width(self.size) else int(self.size, 16)

        if 1 == width:
            self.cs.pci.write_byte(self.bus, self.device, self.function, self.offset, self.value)
        elif 2 == width:
            self.cs.pci.write_word(self.bus, self.device, self.function, self.offset, self.value)
        elif 4 == width:
            self.cs.pci.write_dword(self.bus, self.device, self.function, self.offset, self.value)
        else:
            self.logger.log_error("Width should be one of {}".format(CMD_OPTS_WIDTH))
            return
        self.logger.log("[CHIPSEC] Write 0x{:X} to PCI {:02X}:{:02X}.{:02X} + 0x{:02X}".format(self.value, self.bus, self.device, self.function, self.offset))

    def pci_cmd(self):
        self.logger.log('BDF     | VID:DID   | CMD  | CLS | Sub CLS')
        self.logger.log('------------------------------------------')
        for (b, d, f, vid, did, rid) in self.cs.pci.enumerate_devices():
            dev_cls = self.cs.pci.read_byte(b, d, f, PCI_HDR_CLS_OFF)
            if self.pci_class is not None and (dev_cls != self.pci_class):
                continue
            dev_sub_cls = self.cs.pci.read_byte(b, d, f, PCI_HDR_SUB_CLS_OFF)
            if self.pci_sub_class is not None and (dev_sub_cls != self.pci_sub_class):
                continue
            cmd_reg = self.cs.pci.read_word(b, d, f, PCI_HDR_CMD_OFF)
            if (cmd_reg & self.cmd_mask) == 0:
                continue
            self.logger.log('{:02X}:{:02X}.{:X} | {:04X}:{:04X} | {:04X} | {:02X}  | {:02X}'.format(b, d, f, vid, did, cmd_reg, dev_cls, dev_sub_cls))


commands = {'pci': PCICommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Definition for TPM2.0 commands to use with TPM HAL

TCG PC Client TPM Specification
TCG TPM v2.0 Specification
"""


import struct
from collections import namedtuple
from chipsec.library.logger import logger
from chipsec.library.tpm import tpm_defines


class TPM_RESPONSE_HEADER(namedtuple('TPM_RESPONSE_HEADER', 'ResponseTag DataSize ReturnCode')):
    __slots__ = ()
    def __str__(self) -> str:
        _str = f"""----------------------------------------------------------------
                     TPM response header
----------------------------------------------------------------
   Response Tag: 0x{self.ResponseTag:x}
   Response Size: 0x{self.DataSize:x}
   Response Code: 0x{self.ReturnCode:x}
"""
        _str += '\t'
        try:
            response = tpm_defines.RESPONSE_CODE[self.ReturnCode]
            _str += f'{response[0]}: {response[1]}'
        except:
            _str += 'Invalid return code'

        _str += '\n'
        return _str


def startup(command_argv):
    """
    TPM2_Startup command.
    1: TPM_ST_CLEAR
    2: TPM_ST_STATE
    """
    # session = _read_session(int(command_argv[0]), tpm_defines.TPM_ST_NO_SESSIONS)
    command_format = '=HIIH'
    size = 0x0C000000
    try:
        startup_type = tpm_defines.TPM_SU[int(command_argv[1])]
    except:
        if logger().HAL:
            logger().log_bad("Invalid startup value\n")
        return
    # 8001, 0c000000, 00000144, 0001
    # 0180, 0c000000, 44010000, 0100
    command = struct.pack(command_format, 0x0180, size, 0x44010000, 0x0100)
    return (command, size >> 0x18)

def shutdown(command_argv):
    """
    TPM2_Shutdown command.
    1: TPM_ST_CLEAR
    2: TPM_ST_STATE
    """
    session = _read_session(int(command_argv[0]))
    command_format = '=HIIH'
    size = 0x0C000000
    try:
        shutdown_type = tpm_defines.TPM_SU[int(command_argv[1])]
    except:
        if logger().HAL:
            logger().log_bad("Invalid shutdown value\n")
        return

    command = struct.pack(command_format, session, size, tpm_defines.TPM2_SHUTDOWN_COMMAND, shutdown_type)
    return (command, size >> 0x18)

def selftest(command_argv):
    """
    TPM2_SelfTest command.
    0: YES (if full test to be performed)
    1: NO (if only test of untested functions required)
    """
    session = _read_session(int(command_argv[0]))
    command_format = '=HIIB'
    size = 0x0B000000
    try:
        selftest_yes_no = tpm_defines.TPMI_YES_NO[int(command_argv[1])]
    except:
        if logger().HAL:
            logger().log_bad("Invalid selftest value\n")
        return
    command = struct.pack(command_format, session, size, tpm_defines.TPM2_SELFTEST_COMMAND, selftest_yes_no)
    return (command, size >> 0x18)

def nvread(command_argv):
    """
    TPM2_NV_Read command.
    """
    session = _read_session(int(command_argv[0]), tpm_defines.TPM_ST_SESSIONS)
    command_format = '=HIIIIHH'
    size = 0x0
    try:
        auth = tpm_defines.TPMI_RH_NV_AUTH[int(command_argv[1])]
        index = int(command_argv[2])
        readsize = int(command_argv[3])  # something to do with index_first and index_last...
        readoffset = int(command_argv[4])
    except:
        if logger().HAL:
            logger().log_bad("Invalid NV read value\n")
        return
    command = struct.pack(command_format, session, size, tpm_defines.TPM2_NVREAD_COMMAND, auth, index, readsize, readoffset)
    return (command, size >> 0x18)

def nvwrite(command_argv):
    """
    TPM2_NV_Write command.
    """
    session = _read_session(int(command_argv[0]), tpm_defines.TPM_ST_SESSIONS)
    command_format = '=HIIIIHH'
    size = 0x0
    try:
        auth = tpm_defines.TPMI_RH_NV_AUTH[int(command_argv[1])]
        index = int(command_argv[2])
        writesize = int(command_argv[3])
        writeoffset = int(command_argv[4])
    except:
        if logger().HAL:
            logger().log_bad("Invalid NV read value\n")
        return
    command = struct.pack(command_format, session, size, tpm_defines.TPM2_NVREAD_COMMAND, auth, index, writesize, writeoffset)
    return (command, size >> 0x18)

def pcrread(command_argv):
    """
    TPM2_PCR_Read command.
    """
    # session = _read_session(command_argv[0])
    command_format = '=HIII'
    size = 0x0C000000
    try:
        pcr_selection_in = command_argv[1]
    except:
        if logger().HAL:
            logger().log_bad("Invalid PCR read value\n")
        return
    # 0180, 0c000000, 44010000, 00
    command = struct.pack(command_format, 0x0180, size, 0x7e010000, pcr_selection_in)
    return (command, size >> 0x18)

def nvdefinespace(command_argv):
    """
    TPM2_NV_DefineSpace command.
    """
    session = _read_session(int(command_argv[0]), tpm_defines.TPM_ST_SESSIONS)
    command_format = '=HIIIII'
    size = 0x0
    try:
        provision = tpm_defines.TPMI_RH_NV_AUTH[int(command_argv[1])]
        auth = int(command_argv[2])
        public_info = int(command_argv[3])
    except:
        if logger().HAL:
            logger().log_bad("Invalid NV define space value\n")
        return
    command = struct.pack(command_format, session, size, tpm_defines.TPM2_NVDEFINESPACE_COMMAND, provision, auth, public_info)
    return (command, size >> 0x18)

def nvundefinespace(command_argv):
    """
    TPM2_NV_UndefineSpace command.
    """
    session = _read_session(int(command_argv[0]), tpm_defines.TPM_ST_SESSIONS)
    command_format = '=HIIII'
    size = 0x0
    try:
        provision = tpm_defines.TPMI_RH_NV_AUTH[int(command_argv[1])]
        index = int(command_argv[2])
    except:
        if logger().HAL:
            logger().log_bad("Invalid NV define space value\n")
        return
    command = struct.pack(command_format, session, size, tpm_defines.TPM2_NVUNDEFINESPACE_COMMAND, provision, index)


def _read_session(session_arg, required=None):
    if required and session_arg != required:
        if logger().HAL:
            logger().log_bad('Session value not allowed\n')
        return

    try:
        return tpm_defines.SESSIONS[session_arg]
    except:
        if logger().HAL:
            logger().log_bad('Invalid session value\n')
        return
